from torch_geometric.nn import MessagePassing
from torch_geometric.utils import softmax
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import (
    GCNConv,
    SAGEConv,
    GATConv,
    ARMAConv,
    ChebConv,
    GatedGraphConv,
    SGConv,
)

from typing import Union, Tuple, Optional
from torch_geometric.typing import OptPairTensor, Adj, Size, NoneType, OptTensor

import torch
from torch import Tensor
import torch.nn.functional as F
from torch.nn import Parameter, Linear
from torch_sparse import SparseTensor, set_diag
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.utils import remove_self_loops, add_self_loops, softmax

from torch_geometric.nn.inits import glorot, zeros
import torch
import torch.nn as nn
import torch.nn.functional as F

import torch.nn.functional as F
from torch.nn import Parameter
from torch_geometric.nn.inits import glorot, zeros
from torch_geometric.utils import softmax
from torch_scatter import scatter_add
import numpy as np
from ..operation import GeoLayer


class AggAdd(nn.Module):
    def __init__(
        self, dim, att_head, dropout=0, norm=False, skip_connect=False, *args, **kwargs
    ):
        super(AggAdd, self).__init__()
        self.dropout = dropout
        self.ln_add = nn.BatchNorm1d(dim, track_running_stats=True, affine=True)
        self.norm = norm
        self.skip_connect = skip_connect

    def forward(self, x, edge_index, *args, **kwargs):
        # x=[x_stem,[x_sides]]
        norm = self.norm
        x1, x2, x3 = x[0], x[1][0], x[1][1]
        if norm:
            return self.ln_add(x1 + x2)
        else:
            return x1 + x2


class AggAttn(MessagePassing):
    def __init__(
        self, dim, att_head, dropout=0, norm=False, skip_connect=False, *args, **kwargs
    ):
        super(AggAttn, self).__init__()
        self.dropout = dropout
        self.att_head = att_head
        self.ln_attn = nn.BatchNorm1d(dim, track_running_stats=True, affine=True)
        self.norm = norm
        self.skip_connect = skip_connect

    def __repr__(self) -> str:
        return "AggAttn(att_head={}, dropout={})".format(self.att_head, self.dropout)

    def forward(self, x, edge_index, *args, **kwargs):
        # x=[x_stem,[x_sides]]
        # use dot-product attn
        x1, x2, x3 = x[0], x[1][0], x[1][1]  # q,k,v
        skip_connect, norm = self.skip_connect, self.norm
        if not skip_connect and not norm:
            return self.propagate(edge_index, x1=x1, x2=x2, x3=x3)

        x = self.propagate(edge_index, x1=x1, x2=x2, x3=x3)
        if not norm:
            return x

        if not skip_connect:
            return self.ln_attn(x)
        return self.ln_attn(x + x1)

    def message(self, x2_j, x1_i, x3_j, index, ptr):
        # x1: query, x2: key, x3: value # torch.Size([10556, 64]) ,index torch.Size([10556])
        node, dim = x1_i.size()
        dim_att = dim // self.att_head
        # torch.Size([10556, 8, 8])
        x2_j = x2_j.view(node, self.att_head, dim_att)
        # torch.Size([10556, 8, 8])
        x1_i = x1_i.view(node, self.att_head, dim_att)
        attn = (x2_j * x1_i).sum(dim=-1) / np.sqrt(dim_att)  # torch.Size([10556, 8])
        attn = softmax(attn, index, ptr)  # torch.Size([10556, 8])
        # torch.Size([10556, 8])
        attn = F.dropout(attn, p=self.dropout, training=self.training)
        out = x3_j.view(node, self.att_head, dim_att) * attn.unsqueeze(-1)
        out = out.view(-1, dim)
        return out


class GATConv2(MessagePassing):
    _alpha: OptTensor

    def __init__(
        self,
        in_channels: Union[int, Tuple[int, int]],
        out_channels: int,
        heads: int = 1,
        concat: bool = False,
        negative_slope: float = 0.2,
        dropout: float = 0.0,
        add_self_loops: bool = True,
        bias: bool = True,
        **kwargs
    ):
        super(GATConv, self).__init__(aggr="add", node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.add_self_loops = add_self_loops

        self.lin = Linear(in_channels, heads * out_channels, bias=False)

        self.att = Parameter(torch.Tensor(1, heads, out_channels))

        if bias and concat:
            self.bias = Parameter(torch.Tensor(heads * out_channels))
        elif bias and not concat:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.lin.weight)
        glorot(self.att)
        zeros(self.bias)

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj):
        H, C = self.heads, self.out_channels

        x = self.lin(x).view(-1, H, C)
        alpha = (x * self.att).sum(dim=-1)

        if self.add_self_loops:
            if isinstance(edge_index, Tensor):
                num_nodes = x.size(0)
                edge_index, _ = remove_self_loops(edge_index)
                edge_index, _ = add_self_loops(edge_index, num_nodes=num_nodes)
            elif isinstance(edge_index, SparseTensor):
                edge_index = set_diag(edge_index)

        out = self.propagate(edge_index, x=x, alpha=alpha)

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if self.bias is not None:
            out += self.bias

        return out

    def message(
        self,
        x_j: Tensor,
        alpha_j: Tensor,
        alpha_i: OptTensor,
        index: Tensor,
        ptr: OptTensor,
    ) -> Tensor:
        alpha = alpha_j if alpha_i is None else alpha_j + alpha_i
        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, index, ptr)
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        return x_j * alpha.unsqueeze(-1)

    def __repr__(self):
        return "{}({}, {}, heads={})".format(
            self.__class__.__name__, self.in_channels, self.out_channels, self.heads
        )


class Zero(nn.Module):
    def __init__(self, indim, outdim) -> None:
        super().__init__()
        self.outdim = outdim
        self.zero = nn.Parameter(torch.tensor(0.0), requires_grad=True)

    def forward(self, x, edge_index):
        return torch.zeros(x.size(0), self.outdim).to(x.device) * self.zero


# class Zero(nn.Module):
#     def __init__(self, indim, outdim) -> None:
#         super().__init__()
#         self.outdim = outdim
#         self.ln = nn.Linear(1, 1)

#     def forward(self, x, edge_index):
#         return 0.


class Identity(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x, edge_index):
        return x


class Linear(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.core = nn.Linear(in_dim, out_dim)

    def forward(self, x, *args):
        return self.core(x)


agg_map = {
    "add": lambda dim, att_head=None, dropout=0, norm=False, skip_connect=False: AggAdd(
        dim, att_head, dropout, norm, skip_connect
    ),
    "attn": lambda dim, att_head=None, dropout=0, norm=False, skip_connect=False: AggAttn(
        dim, att_head, dropout, norm, skip_connect
    ),
}



from torch import nn
import torch


class PathSamplingLayerChoice(nn.Module):
    """
    Mixed module, in which fprop is decided by exactly one or multiple (sampled) module.
    If multiple module is selected, the result will be sumed and returned.

    Attributes
    ----------
    sampled : int or list of int
        Sampled module indices.
    mask : tensor
        A multi-hot bool 1D-tensor representing the sampled mask.
    """

    def __init__(self, layer_choice):
        super(PathSamplingLayerChoice, self).__init__()
        self.op_names = []
        for name, module in layer_choice.named_children():
            self.add_module(name, module)
            self.op_names.append(name)
        assert self.op_names, "There has to be at least one op to choose from."
        self.sampled = None  # sampled can be either a list of indices or an index

    def forward(self, *args, **kwargs):
        assert (
            self.sampled is not None
        ), "At least one path needs to be sampled before fprop."
        if isinstance(self.sampled, list):
            return sum(
                [getattr(self, self.op_names[i])(*args, **kwargs) for i in self.sampled]
            )  # pylint: disable=not-an-iterable
        else:
            return getattr(self, self.op_names[self.sampled])(
                *args, **kwargs
            )  # pylint: disable=invalid-sequence-index

    def sampled_choices(self):
        if self.sampled is None:
            return []
        elif isinstance(self.sampled, list):
            return [
                getattr(self, self.op_names[i]) for i in self.sampled
            ]  # pylint: disable=not-an-iterable
        else:
            return [
                getattr(self, self.op_names[self.sampled])
            ]  # pylint: disable=invalid-sequence-index

    def __len__(self):
        return len(self.op_names)

    @property
    def mask(self):
        return _get_mask(self.sampled, len(self))

    def __repr__(self):
        return (
            f"PathSamplingLayerChoice(op_names={self.op_names}, chosen={self.sampled})"
        )


class PathSamplingInputChoice(nn.Module):
    """
    Mixed input. Take a list of tensor as input, select some of them and return the sum.

    Attributes
    ----------
    sampled : int or list of int
        Sampled module indices.
    mask : tensor
        A multi-hot bool 1D-tensor representing the sampled mask.
    """

    def __init__(self, input_choice):
        super(PathSamplingInputChoice, self).__init__()
        self.n_candidates = input_choice.n_candidates
        self.n_chosen = input_choice.n_chosen
        self.sampled = None

    def forward(self, input_tensors):
        if isinstance(self.sampled, list):
            return sum(
                [input_tensors[t] for t in self.sampled]
            )  # pylint: disable=not-an-iterable
        else:
            return input_tensors[self.sampled]

    def __len__(self):
        return self.n_candidates

    @property
    def mask(self):
        return _get_mask(self.sampled, len(self))

    def __repr__(self):
        return f"PathSamplingInputChoice(n_candidates={self.n_candidates}, chosen={self.sampled})"



"""
Base estimator of NAS
"""

from abc import abstractmethod
from ..space import BaseSpace
from typing import Tuple
import torch.nn.functional as F
import torch


class BaseEstimator:
    """
    The estimator of NAS model.

    Parameters
    ----------
    loss_f: callable
        Default loss function for evaluation

    evaluation: list of autogl.module.train.evaluation.Evaluation
        Default evaluation metric
    """

    def __init__(self, loss_f, evaluation):
        """
        Initial estimator

        Parameters
        ----------
        loss_f: None || function(BaseSpace, dataset, mask)->torch.Tensor
            Get the loss function, should be differentiable

        evaluation: None || function(BaseSpace, dataset, mask)->float
            Get an evaluation function

        """
        self.loss_f = loss_f
        self.evaluation = evaluation

    def setLossFunction(self, loss_f: str):
        self.loss_f = loss_f

    def setEvaluation(self, evaluation):
        self.evaluation = evaluation

    @abstractmethod
    def infer(
        self, model: BaseSpace, dataset, mask="train", *args, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Calculate the loss and metrics of given model on given dataset using
        specified masks.

        Parameters
        ----------
        model: autogl.module.nas.space.BaseSpace
            The model in space.

        dataset: any
            The dataset to perform infer

        mask: str
            The mask to evalute on dataset

        Return
        ------
        metrics: list of float
            the metrics on given datasets.
        loss: torch.Tensor
            the loss on given datasets. Note that loss should be differentiable.
        """
        raise NotImplementedError()



import torch
from torch import nn
import torch_geometric.nn as pygnn
import random

import dgl.function as fn
from dgl.ops import edge_softmax
from dgl.utils import expand_as_pair
import dgl.nn.pytorch as dglnn

class LinearConv(torch.nn.Linear):
    def forward(self, x, *args, **kwargs):
        if isinstance(x, tuple): x = x[1]
        return super().forward(x)

class IdentityConv(torch.nn.Identity):
    def forward(self, x, *args, **kwargs):
        if isinstance(x, tuple): x = x[1]
        return super().forward(x)
    
    def reset_parameters(self):
        return

class ZeroConv(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.dummy = torch.nn.Parameter(torch.tensor([0.]), True)
    
    def forward(self, x, *args, **kwargs):
        if isinstance(x, tuple): x = x[1]
        return torch.zeros_like(x).to(x.device) + 0.0 * self.dummy
    
    def reset_parameters(self):
        return

class Wrap(torch.nn.Module):
    def __init__(self, indim, outdim, module) -> None:
        super().__init__()
        if indim != outdim:
            self.map = torch.nn.Linear(indim, outdim)
        else:
            self.map = torch.nn.Identity()
        self.core = module
    
    def forward(self, x, edge_index):
        x = self.map(x)
        return self.core(x, edge_index)
    
    def reset_parameters(self):
        if hasattr(self.map, 'reset_parameters'):
            self.map.reset_parameters()
        self.core.reset_parameters()

class MLP(torch.nn.Module):
    r"""A multi-layer perception (MLP) model.

    Args:
        channel_list (List[int]): List of input, intermediate and output
            channels. :obj:`len(channel_list) - 1` denotes the number of layers
            of the MLP.
        dropout (float, optional): Dropout probability of each hidden
            embedding. (default: :obj:`0.`)
        batch_norm (bool, optional): If set to :obj:`False`, will not make use
            of batch normalization. (default: :obj:`True`)
        relu_first (bool, optional): If set to :obj:`True`, ReLU activation is
            applied before batch normalization. (default: :obj:`False`)
    """
    def __init__(self, channel_list, dropout: float = 0.,
                 batch_norm: bool = True, relu_first: bool = False):
        super().__init__()
        assert len(channel_list) >= 2
        self.channel_list = channel_list
        self.dropout = dropout
        self.relu_first = relu_first

        self.lins = torch.nn.ModuleList()
        for dims in zip(channel_list[:-1], channel_list[1:]):
            self.lins.append(torch.nn.Linear(*dims))

        self.norms = torch.nn.ModuleList()
        for dim in zip(channel_list[1:-1]):
            self.norms.append(torch.nn.BatchNorm1d(dim) if batch_norm else torch.nn.Identity())

        self.reset_parameters()

    def reset_parameters(self):
        for lin in self.lins:
            lin.reset_parameters()
        for norm in self.norms:
            if hasattr(norm, 'reset_parameters'):
                norm.reset_parameters()


    def forward(self, x):
        """"""
        x = self.lins[0](x)
        for lin, norm in zip(self.lins[1:], self.norms):
            if self.relu_first:
                x = x.relu_()
            x = norm(x)
            if not self.relu_first:
                x = x.relu_()
            x = torch.nn.functional.dropout(x, p=self.dropout, training=self.training)
            x = lin.forward(x)
        return x


    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({str(self.channel_list)[1:-1]})'

# define operators
def op(name, num_input, num_output=None, dropout=0.5, **kwargs):
    num_output = num_output or num_input
    if name == "gcn": # currently no support for neighbor sampling
        return pygnn.GCNConv(num_input, num_output)
    if name == "gat1":
        return pygnn.GATConv(num_input, num_output // 1, heads=1)
    if name == "gat2":
        return pygnn.GATConv(num_input, num_output // 2, heads=2)
    if name == "gat4":
        return pygnn.GATConv(num_input, num_output // 4, heads=4)
    if name == "gat8":
        return pygnn.GATConv(num_input, num_output // 8, heads=8)
    if name == "gat16":
        return pygnn.GATConv(num_input, num_output // 16, heads=16)
    if name == "sage":
        return pygnn.SAGEConv(num_input, num_output)
    if name == "cheb": # currently no support for neighbor sampling
        return pygnn.ChebConv(num_input, num_output, 2)
    if name == "graph":
        return pygnn.GraphConv(num_input, num_output, aggr='mean')
    if name == "transformer":
        return pygnn.TransformerConv(num_input, num_output)
    if name == "gin":
        return pygnn.GINConv(nn.Sequential(
            nn.Linear(num_input, num_output),
            nn.ReLU(),
            nn.Linear(num_output, num_output)
        ), aggr='mean')
    if name == 'ginv2':
        return pygnn.GINConv(
            nn.Sequential(
                nn.BatchNorm1d(num_input),
                nn.Linear(num_input, num_output),
                nn.ReLU(),
                nn.Linear(num_output, num_output)
            )
        )
    if name == 'ginv3':
        return pygnn.GINConv(MLP([num_input, num_output, num_output]))
    if name == "resggraph":
        return pygnn.ResGatedGraphConv(num_input, num_output)
    if name == "gatv2":
        return pygnn.GATv2Conv(num_input, num_output)
    if name == "cg":
        return pygnn.CGConv(num_input, num_output)
    if name == "feast":
        return pygnn.FeaStConv(num_input, num_output)
    if name == "le":
        return pygnn.LEConv(num_input, num_output)
    if name == "film":
        return pygnn.FiLMConv(num_input, num_output)
    if name == "linear":
        return LinearConv(num_input, num_output)
    if name == "identity":
        return IdentityConv()
    if name == "zero":
        return ZeroConv()
    if name == 'sgc':
        return pygnn.SGConv(num_input, num_output)
    if name == 'appnp':
        return Wrap(num_input, num_output, pygnn.APPNP(10, 0.2))

def uniform_sample(layers, space):
    if isinstance(space[0], list):
        assert len(space) == layers
        return [random.choice(s) for s in space]
    return [random.choice(space) for _ in range(layers)]

def uniform_samples(layers, space, number):
    archs = []
    while len(archs) < number:
        arch = uniform_sample(layers, space)
        if arch not in archs: archs.append(arch)
    return archs

def mutate(arch, space, ratio="auto"):
    if ratio == "auto":
        ratio = 1 - (0.5 ** (1 / len(arch)))
    new_arch = []
    for i, a in enumerate(arch):
        if random.random() < ratio:
            # mutate
            if isinstance(space[0], list):
                new_arch.append(random.choice(space[i]))
            else:
                new_arch.append(random.choice(space))
        else:
            new_arch.append(a)
    if new_arch == arch:
        return mutate(arch, space, ratio)
    return new_arch

# borrowed from baseline
class GATConvDGL(nn.Module):
    def __init__(
        self,
        node_feats,
        edge_feats,
        out_feats,
        n_heads=1,
        attn_drop=0.0,
        edge_drop=0.0,
        negative_slope=0.2,
        residual=True,
        activation=None,
        use_attn_dst=True,
        allow_zero_in_degree=True,
        use_symmetric_norm=False,
    ):
        super(GATConvDGL, self).__init__()
        self._n_heads = n_heads
        self._in_src_feats, self._in_dst_feats = expand_as_pair(node_feats)
        self._out_feats = out_feats
        self._allow_zero_in_degree = allow_zero_in_degree
        self._use_symmetric_norm = use_symmetric_norm

        # feat fc
        self.src_fc = nn.Linear(self._in_src_feats, out_feats * n_heads, bias=False)
        if residual:
            self.dst_fc = nn.Linear(self._in_src_feats, out_feats * n_heads)
            self.bias = None
        else:
            self.dst_fc = None
            self.bias = nn.Parameter(out_feats * n_heads)

        # attn fc
        self.attn_src_fc = nn.Linear(self._in_src_feats, n_heads, bias=False)
        if use_attn_dst:
            self.attn_dst_fc = nn.Linear(self._in_src_feats, n_heads, bias=False)
        else:
            self.attn_dst_fc = None
        if edge_feats > 0:
            self.attn_edge_fc = nn.Linear(edge_feats, n_heads, bias=False)
        else:
            self.attn_edge_fc = None

        self.attn_drop = nn.Dropout(attn_drop)
        self.edge_drop = edge_drop
        self.leaky_relu = nn.LeakyReLU(negative_slope, inplace=True)
        self.activation = activation

        self.reset_parameters()

    def reset_parameters(self):
        gain = nn.init.calculate_gain("relu")
        nn.init.xavier_normal_(self.src_fc.weight, gain=gain)
        if self.dst_fc is not None:
            nn.init.xavier_normal_(self.dst_fc.weight, gain=gain)

        nn.init.xavier_normal_(self.attn_src_fc.weight, gain=gain)
        if self.attn_dst_fc is not None:
            nn.init.xavier_normal_(self.attn_dst_fc.weight, gain=gain)
        if self.attn_edge_fc is not None:
            nn.init.xavier_normal_(self.attn_edge_fc.weight, gain=gain)

        if self.bias is not None:
            nn.init.zeros_(self.bias)

    def set_allow_zero_in_degree(self, set_value):
        self._allow_zero_in_degree = set_value

    def forward(self, graph, feat_src, feat_edge=None):
        with graph.local_scope():
            if not self._allow_zero_in_degree:
                if (graph.in_degrees() == 0).any():
                    assert False

            if graph.is_block:
                feat_dst = feat_src[: graph.number_of_dst_nodes()]
            else:
                feat_dst = feat_src

            if self._use_symmetric_norm:
                degs = graph.srcdata["deg"]
                # degs = graph.out_degrees().float().clamp(min=1)
                norm = torch.pow(degs, -0.5)
                shp = norm.shape + (1,) * (feat_src.dim() - 1)
                norm = torch.reshape(norm, shp)
                feat_src = feat_src * norm

            feat_src_fc = self.src_fc(feat_src).view(-1, self._n_heads, self._out_feats)
            feat_dst_fc = self.dst_fc(feat_dst).view(-1, self._n_heads, self._out_feats)
            attn_src = self.attn_src_fc(feat_src).view(-1, self._n_heads, 1)

            # NOTE: GAT paper uses "first concatenation then linear projection"
            # to compute attention scores, while ours is "first projection then
            # addition", the two approaches are mathematically equivalent:
            # We decompose the weight vector a mentioned in the paper into
            # [a_l || a_r], then
            # a^T [Wh_i || Wh_j] = a_l Wh_i + a_r Wh_j
            # Our implementation is much efficient because we do not need to
            # save [Wh_i || Wh_j] on edges, which is not memory-efficient. Plus,
            # addition could be optimized with DGL's built-in function u_add_v,
            # which further speeds up computation and saves memory footprint.
            graph.srcdata.update({"feat_src_fc": feat_src_fc, "attn_src": attn_src})

            if self.attn_dst_fc is not None:
                attn_dst = self.attn_dst_fc(feat_dst).view(-1, self._n_heads, 1)
                graph.dstdata.update({"attn_dst": attn_dst})
                graph.apply_edges(fn.u_add_v("attn_src", "attn_dst", "attn_node"))
            else:
                graph.apply_edges(fn.copy_u("attn_src", "attn_node"))

            e = graph.edata["attn_node"]
            if feat_edge is not None:
                attn_edge = self.attn_edge_fc(feat_edge).view(-1, self._n_heads, 1)
                graph.edata.update({"attn_edge": attn_edge})
                e += graph.edata["attn_edge"]
            e = self.leaky_relu(e)

            if self.training and self.edge_drop > 0:
                perm = torch.randperm(graph.number_of_edges(), device=e.device)
                bound = int(graph.number_of_edges() * self.edge_drop)
                eids = perm[bound:]
                graph.edata["a"] = torch.zeros_like(e)
                graph.edata["a"][eids] = self.attn_drop(edge_softmax(graph, e[eids], eids=eids))
            else:
                graph.edata["a"] = self.attn_drop(edge_softmax(graph, e))

            # message passing
            graph.update_all(fn.u_mul_e("feat_src_fc", "a", "m"), fn.sum("m", "feat_src_fc"))

            rst = graph.dstdata["feat_src_fc"]

            if self._use_symmetric_norm:
                degs = graph.dstdata["deg"]
                # degs = graph.in_degrees().float().clamp(min=1)
                norm = torch.pow(degs, 0.5)
                shp = norm.shape + (1,) * (feat_dst.dim())
                norm = torch.reshape(norm, shp)
                rst = rst * norm

            # residual
            if self.dst_fc is not None:
                rst += feat_dst_fc
            else:
                rst += self.bias

            # activation
            if self.activation is not None:
                rst = self.activation(rst, inplace=True)

            return rst.flatten(1, -1)

class EdgeConv(nn.Module):
    def __init__(self, conv) -> None:
        super().__init__()
        self.conv = conv
    
    def forward(self, g, node_feat, edge_feat):
        return self.conv(g, node_feat)


class Linear(nn.Module):
    def __init__(self, in_dim, out_dim) -> None:
        super().__init__()
        self.core = nn.Linear(in_dim, out_dim)
    
    def forward(self, g, node_feat, edge_feat):
        return self.core(node_feat)


def op_dgl(name, in_dim, out_dim, edge_feat, **kwargs):
    if name == 'gat':
        return GATConvDGL(
            in_dim,
            edge_feat,
            out_dim // kwargs['n_heads'],
            n_heads=kwargs['n_heads'],
            attn_drop=kwargs['attn_drop'],
            edge_drop=kwargs['edge_drop'],
            use_attn_dst=kwargs['use_attn_dst'],
            allow_zero_in_degree=kwargs['allow_zero_in_degree'],
            use_symmetric_norm=False
        )
    if name == 'gcn':
        return EdgeConv(dglnn.GraphConv(in_dim, out_dim, allow_zero_in_degree=kwargs['allow_zero_in_degree']))
    if name == 'tag':
        return EdgeConv(dglnn.TAGConv(in_dim, out_dim))
    if name == 'edge':
        return EdgeConv(dglnn.EdgeConv(in_dim, out_dim, allow_zero_in_degree=kwargs['allow_zero_in_degree']))
    if name == 'sage':
        return EdgeConv(dglnn.SAGEConv(in_dim, out_dim, 'mean'))
    if name == 'sg':
        return EdgeConv(dglnn.SGConv(in_dim, out_dim, allow_zero_in_degree=kwargs['allow_zero_in_degree']))
    if name == 'appnp':
        return EdgeConv(dglnn.APPNPConv(10, 0.1))
    if name == 'gin':
        return EdgeConv(dglnn.GINConv(nn.Sequential(
            nn.Linear(in_dim, out_dim),
            nn.ReLU(),
            nn.Linear(out_dim, out_dim)
        ), 'max'))
    if name == 'gmm':
        return dglnn.GMMConv(in_dim, out_dim, edge_feat, 5)
    if name == 'cheb':
        return EdgeConv(dglnn.ChebConv(in_dim, out_dim, 2))
    if name == 'agnn':
        return EdgeConv(dglnn.AGNNConv(allow_zero_in_degree=kwargs['allow_zero_in_degree']))
    if name == 'cf':
        return dglnn.CFConv(in_dim, edge_feat, out_dim, out_dim)
    if name == 'linear':
        return Linear(in_dim, out_dim)



from .nni import (
    apply_fixed_architecture,
    OrderedMutable,
    OrderedLayerChoice,
    OrderedInputChoice,
    get_module_order,
    sort_replaced_module,
    replace_layer_choice,
    replace_input_choice,
)
from .nni import PathSamplingLayerChoice, PathSamplingInputChoice
from .nni import DartsLayerChoice, DartsInputChoice

from .base import BaseSpace
from .single_path import SinglePathNodeClassificationSpace
from .gasso import GassoSpace
from .graces import GracesSpace
from .gauss import GaussSpace
from .autogt import AutogtSpace
from .graph_nas import GraphNasNodeClassificationSpace
from .graph_nas_macro import GraphNasMacroNodeClassificationSpace
from .autoattend import AutoAttendNodeClassificationSpace



import re
import inspect
import pyparsing as pp
from itertools import product
from collections import OrderedDict
from typing import Callable, Tuple, Dict, List


def split_types_repr(types_repr: str) -> List[str]:
    out = []
    i = depth = 0
    for j, char in enumerate(types_repr):
        if char == "[":
            depth += 1
        elif char == "]":
            depth -= 1
        elif char == "," and depth == 0:
            out.append(types_repr[i:j].strip())
            i = j + 1
    out.append(types_repr[i:].strip())
    return out


def sanitize(type_repr: str):
    type_repr = re.sub(r"<class \'(.*)\'>", r"\1", type_repr)
    type_repr = type_repr.replace("typing.", "")
    type_repr = type_repr.replace("torch_sparse.tensor.", "")
    type_repr = type_repr.replace("Adj", "Union[Tensor, SparseTensor]")

    # Replace `Union[..., NoneType]` by `Optional[...]`.
    sexp = pp.nestedExpr(opener="[", closer="]")
    tree = sexp.parseString(f'[{type_repr.replace(",", " ")}]').asList()[0]

    def union_to_optional_(tree):
        for i in range(len(tree)):
            e, n = tree[i], tree[i + 1] if i + 1 < len(tree) else []
            if e == "Union" and n[-1] == "NoneType":
                tree[i] = "Optional"
                tree[i + 1] = tree[i + 1][:-1]
            elif e == "Union" and "NoneType" in n:
                idx = n.index("NoneType")
                n[idx] = [n[idx - 1]]
                n[idx - 1] = "Optional"
            elif isinstance(e, list):
                tree[i] = union_to_optional_(e)
        return tree

    tree = union_to_optional_(tree)
    type_repr = re.sub(r"\'|\"", "", str(tree)[1:-1]).replace(", [", "[")

    return type_repr


def param_type_repr(param) -> str:
    if param.annotation is inspect.Parameter.empty:
        return "torch.Tensor"
    return sanitize(re.split(r":|=".strip(), str(param))[1])


def return_type_repr(signature) -> str:
    return_type = signature.return_annotation
    if return_type is inspect.Parameter.empty:
        return "torch.Tensor"
    elif str(return_type)[:6] != "<class":
        return sanitize(str(return_type))
    elif return_type.__module__ == "builtins":
        return return_type.__name__
    else:
        return f"{return_type.__module__}.{return_type.__name__}"


def parse_types(func: Callable) -> List[Tuple[Dict[str, str], str]]:
    source = inspect.getsource(func)
    signature = inspect.signature(func)

    # Parse `# type: (...) -> ...` annotation. Note that it is allowed to pass
    # multiple `# type:` annotations in `forward()`.
    iterator = re.finditer(r"#\s*type:\s*\((.*)\)\s*->\s*(.*)\s*\n", source)
    matches = list(iterator)

    if len(matches) > 0:
        out = []
        args = list(signature.parameters.keys())
        for match in matches:
            arg_types_repr, return_type = match.groups()
            arg_types = split_types_repr(arg_types_repr)
            arg_types = OrderedDict((k, v) for k, v in zip(args, arg_types))
            return_type = return_type.split("#")[0].strip()
            out.append((arg_types, return_type))
        return out

    # Alternatively, parse annotations using the inspected signature.
    else:
        ps = signature.parameters
        arg_types = OrderedDict((k, param_type_repr(v)) for k, v in ps.items())
        return [(arg_types, return_type_repr(signature))]


def resolve_types(
    arg_types: Dict[str, str], return_type_repr: str
) -> List[Tuple[List[str], str]]:
    out = []
    for type_repr in arg_types.values():
        if type_repr[:5] == "Union":
            out.append(split_types_repr(type_repr[6:-1]))
        else:
            out.append([type_repr])
    return [(x, return_type_repr) for x in product(*out)]



# "Graph Neural Architecture Search Under Distribution Shifts" ICML 22'

import torch
from tqdm import trange

from ..estimator.base import BaseEstimator
from ..space import BaseSpace
from .base import BaseNAS


class Graces(BaseNAS):
    """
    GRACES trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    device : str or torch.device
        The device of the whole process
    """

    def __init__(
        self,
        num_epochs=250,
        device="auto",
        disable_progress=False,
        args=None,
    ):
        super().__init__(device=device)
        self.num_epochs = num_epochs
        self.disable_progress = disable_progress
        self.args = args

    def train_graph(
        self,
        model_optimizer,
        arch_optimizer,
        gnn0_optimizer,
        eta,
    ):
        self.space.train()

        for id, train_data in enumerate(self.train_loader):
            model_optimizer.zero_grad()
            arch_optimizer.zero_grad()
            gnn0_optimizer.zero_grad()
            train_data = train_data.to(self.device)

            output0, output, cosloss, sslout = self.space(train_data)
            output0 = output0.to(self.device)
            output = output.to(self.device)

            is_labeled = train_data.y == train_data.y
            error_loss0 = self.criterion(
                output0.to(torch.float32)[is_labeled],
                train_data.y.to(torch.float32)[is_labeled],
            )
            error_loss = self.criterion(
                output.to(torch.float32)[is_labeled],
                train_data.y.to(torch.float32)[is_labeled],
            )

            ssltarget = train_data.deratio.view(-1, 3)
            ssllossfun = torch.nn.L1Loss()
            sslloss = ssllossfun(sslout, ssltarget)

            my_loss = (1 - eta) * (
                error_loss0 + self.args.gamma * sslloss + self.args.beta * cosloss
            ) + eta * error_loss

            my_loss.backward()
            model_optimizer.step()
            gnn0_optimizer.step()
            arch_optimizer.step()

    def _infer(self, mask="train"):
        if mask == "train":
            dataloader = self.train_loader
        elif mask == "val":
            dataloader = self.val_loader
        elif mask == "test":
            dataloader = self.test_loader
        metric, loss = self.estimator.infer(self.space, dataloader)
        return metric, loss

    def prepare(self, data):
        """
        data : list of data objects.
            [dataset, train_dataset, val_dataset, test_dataset, train_loader, val_loader, test_loader]
        """
        self.train_loader = data[4]
        self.val_loader = data[5]
        self.test_loader = data[6]

    def fit(self):
        optimizer = torch.optim.Adam(
            self.space.supernet.parameters(),
            self.args.learning_rate,
            weight_decay=self.args.weight_decay,
        )
        arch_optimizer = torch.optim.Adam(
            self.space.ag.parameters(),
            self.args.arch_learning_rate,
            weight_decay=self.args.arch_weight_decay,
        )
        gnn0_optimizer = torch.optim.Adam(
            self.space.supernet0.parameters(),
            self.args.gnn0_learning_rate,
            weight_decay=self.args.gnn0_weight_decay,
        )
        scheduler_arch = torch.optim.lr_scheduler.CosineAnnealingLR(
            arch_optimizer,
            float(self.num_epochs),
            eta_min=self.args.arch_learning_rate_min,
        )
        scheduler_gnn0 = torch.optim.lr_scheduler.CosineAnnealingLR(
            gnn0_optimizer,
            float(self.num_epochs),
            eta_min=self.args.gnn0_learning_rate_min,
        )

        self.criterion = torch.nn.BCEWithLogitsLoss()
        eta = self.args.eta
        best_performance = 0
        min_val_loss = float("inf")

        with trange(self.num_epochs, disable=self.disable_progress) as bar:
            for epoch in bar:
                """
                space training
                """
                self.space.train()
                eta = (
                    self.args.eta_max - self.args.eta
                ) * epoch / self.num_epochs + self.args.eta
                optimizer.zero_grad()
                arch_optimizer.zero_grad()
                gnn0_optimizer.zero_grad()

                self.train_graph(
                    optimizer,
                    arch_optimizer,
                    gnn0_optimizer,
                    eta,
                )
                scheduler_arch.step()
                scheduler_gnn0.step()

                """
                space evaluation
                """
                self.space.eval()
                train_metric, train_loss = self._infer("train")
                val_metric, val_loss = self._infer("val")
                # test_metric, test_loss = self._infer("test")

                if min_val_loss > val_loss:
                    min_val_loss, best_performance = val_loss, val_metric["auc"]
                    self.space.keep_prediction()

                bar.set_postfix(
                    {
                        "train_auc": train_metric["auc"],
                        "val_auc": val_metric["auc"],
                        # "test_auc": test_metric["auc"],
                    }
                )

        return best_performance, min_val_loss

    def search(self, space: BaseSpace, dataset, estimator: BaseEstimator):
        self.estimator = estimator
        self.space = space.to(self.device)
        self.prepare(dataset)
        perf, val_loss = self.fit()
        return space.parse_model(None)



import os
import re
import inspect
import os.path as osp
from uuid import uuid1
from itertools import chain
from inspect import Parameter
from typing import List, Optional, Set
from torch_geometric.typing import Adj, Size

import torch
from torch import Tensor
from jinja2 import Template
from torch_sparse import SparseTensor
from torch_scatter import gather_csr, scatter, segment_csr

from .utils.helpers import expand_left
from .utils.jit import class_from_module_repr
from .utils.typing import sanitize, split_types_repr, parse_types, resolve_types
from .utils.inspector import Inspector, func_header_repr, func_body_repr


class MessagePassing(torch.nn.Module):
    r"""Base class for creating message passing layers of the form

    .. math::
        \mathbf{x}_i^{\prime} = \gamma_{\mathbf{\Theta}} \left( \mathbf{x}_i,
        \square_{j \in \mathcal{N}(i)} \, \phi_{\mathbf{\Theta}}
        \left(\mathbf{x}_i, \mathbf{x}_j,\mathbf{e}_{j,i}\right) \right),

    where :math:`\square` denotes a differentiable, permutation invariant
    function, *e.g.*, sum, mean or max, and :math:`\gamma_{\mathbf{\Theta}}`
    and :math:`\phi_{\mathbf{\Theta}}` denote differentiable functions such as
    MLPs.
    See `here <https://pytorch-geometric.readthedocs.io/en/latest/notes/
    create_gnn.html>`__ for the accompanying tutorial.

    Args:
        aggr (string, optional): The aggregation scheme to use
            (:obj:`"add"`, :obj:`"mean"`, :obj:`"max"` or :obj:`None`).
            (default: :obj:`"add"`)
        flow (string, optional): The flow direction of message passing
            (:obj:`"source_to_target"` or :obj:`"target_to_source"`).
            (default: :obj:`"source_to_target"`)
        node_dim (int, optional): The axis along which to propagate.
            (default: :obj:`-2`)
    """

    special_args: Set[str] = {
        "edge_index",
        "adj_t",
        "edge_index_i",
        "edge_index_j",
        "size",
        "size_i",
        "size_j",
        "ptr",
        "index",
        "dim_size",
    }

    def __init__(
        self,
        aggr: Optional[str] = "add",
        flow: str = "source_to_target",
        node_dim: int = -2,
    ):

        super(MessagePassing, self).__init__()

        self.aggr = aggr
        assert self.aggr in ["add", "mean", "max", None]

        self.flow = flow
        assert self.flow in ["source_to_target", "target_to_source"]

        self.node_dim = node_dim

        self.inspector = Inspector(self)
        self.inspector.inspect(self.message)
        self.inspector.inspect(self.aggregate, pop_first=True)
        self.inspector.inspect(self.message_and_aggregate, pop_first=True)
        self.inspector.inspect(self.update, pop_first=True)

        self.__user_args__ = self.inspector.keys(
            ["message", "aggregate", "update"]
        ).difference(self.special_args)
        self.__fused_user_args__ = self.inspector.keys(
            ["message_and_aggregate", "update"]
        ).difference(self.special_args)

        # Support for "fused" message passing.
        self.fuse = self.inspector.implements("message_and_aggregate")

        # Support for GNNExplainer.
        self.__explain__ = False
        self.__edge_mask__ = None

    def __check_input__(self, edge_index, size):
        the_size: List[Optional[int]] = [None, None]

        if isinstance(edge_index, Tensor):
            assert edge_index.dtype == torch.long
            assert edge_index.dim() == 2
            assert edge_index.size(0) == 2
            if size is not None:
                the_size[0] = size[0]
                the_size[1] = size[1]
            return the_size

        elif isinstance(edge_index, SparseTensor):
            if self.flow == "target_to_source":
                raise ValueError(
                    (
                        'Flow direction "target_to_source" is invalid for '
                        "message propagation via `torch_sparse.SparseTensor`. If "
                        "you really want to make use of a reverse message "
                        "passing flow, pass in the transposed sparse tensor to "
                        "the message passing module, e.g., `adj_t.t()`."
                    )
                )
            the_size[0] = edge_index.sparse_size(1)
            the_size[1] = edge_index.sparse_size(0)
            return the_size

        raise ValueError(
            (
                "`MessagePassing.propagate` only supports `torch.LongTensor` of "
                "shape `[2, num_messages]` or `torch_sparse.SparseTensor` for "
                "argument `edge_index`."
            )
        )

    def __set_size__(self, size: List[Optional[int]], dim: int, src: Tensor):
        the_size = size[dim]
        if the_size is None:
            size[dim] = src.size(self.node_dim)
        elif the_size != src.size(self.node_dim):
            raise ValueError(
                (
                    f"Encountered tensor with size {src.size(self.node_dim)} in "
                    f"dimension {self.node_dim}, but expected size {the_size}."
                )
            )

    def __lift__(self, src, edge_index, dim):
        if isinstance(edge_index, Tensor):
            index = edge_index[dim]
            return src.index_select(self.node_dim, index)
        elif isinstance(edge_index, SparseTensor):
            if dim == 1:
                rowptr = edge_index.storage.rowptr()
                rowptr = expand_left(rowptr, dim=self.node_dim, dims=src.dim())
                return gather_csr(src, rowptr)
            elif dim == 0:
                col = edge_index.storage.col()
                return src.index_select(self.node_dim, col)
        raise ValueError

    def __collect__(self, args, edge_index, size, kwargs):
        i, j = (1, 0) if self.flow == "source_to_target" else (0, 1)

        out = {}
        for arg in args:
            if arg[-2:] not in ["_i", "_j"]:
                out[arg] = kwargs.get(arg, Parameter.empty)
            else:
                dim = 0 if arg[-2:] == "_j" else 1
                data = kwargs.get(arg[:-2], Parameter.empty)

                if isinstance(data, (tuple, list)):
                    assert len(data) == 2
                    if isinstance(data[1 - dim], Tensor):
                        self.__set_size__(size, 1 - dim, data[1 - dim])
                    data = data[dim]

                if isinstance(data, Tensor):
                    self.__set_size__(size, dim, data)
                    data = self.__lift__(data, edge_index, j if arg[-2:] == "_j" else i)

                out[arg] = data

        if isinstance(edge_index, Tensor):
            out["adj_t"] = None
            out["edge_index"] = edge_index
            out["edge_index_i"] = edge_index[i]
            out["edge_index_j"] = edge_index[j]
            out["ptr"] = None
        elif isinstance(edge_index, SparseTensor):
            out["adj_t"] = edge_index
            out["edge_index"] = None
            out["edge_index_i"] = edge_index.storage.row()
            out["edge_index_j"] = edge_index.storage.col()
            out["ptr"] = edge_index.storage.rowptr()
            out["edge_weight"] = edge_index.storage.value()
            out["edge_attr"] = edge_index.storage.value()
            out["edge_type"] = edge_index.storage.value()

        out["index"] = out["edge_index_i"]
        out["size"] = size
        out["size_i"] = size[1] or size[0]
        out["size_j"] = size[0] or size[1]
        out["dim_size"] = out["size_i"]

        return out

    def propagate(self, edge_index: Adj, size: Size = None, **kwargs):
        r"""The initial call to start propagating messages.

        Args:
            edge_index (Tensor or SparseTensor): A :obj:`torch.LongTensor` or a
                :obj:`torch_sparse.SparseTensor` that defines the underlying
                graph connectivity/message passing flow.
                :obj:`edge_index` holds the indices of a general (sparse)
                assignment matrix of shape :obj:`[N, M]`.
                If :obj:`edge_index` is of type :obj:`torch.LongTensor`, its
                shape must be defined as :obj:`[2, num_messages]`, where
                messages from nodes in :obj:`edge_index[0]` are sent to
                nodes in :obj:`edge_index[1]`
                (in case :obj:`flow="source_to_target"`).
                If :obj:`edge_index` is of type
                :obj:`torch_sparse.SparseTensor`, its sparse indices
                :obj:`(row, col)` should relate to :obj:`row = edge_index[1]`
                and :obj:`col = edge_index[0]`.
                The major difference between both formats is that we need to
                input the *transposed* sparse adjacency matrix into
                :func:`propagate`.
            size (tuple, optional): The size :obj:`(N, M)` of the assignment
                matrix in case :obj:`edge_index` is a :obj:`LongTensor`.
                If set to :obj:`None`, the size will be automatically inferred
                and assumed to be quadratic.
                This argument is ignored in case :obj:`edge_index` is a
                :obj:`torch_sparse.SparseTensor`. (default: :obj:`None`)
            **kwargs: Any additional data which is needed to construct and
                aggregate messages, and to update node embeddings.
        """
        size = self.__check_input__(edge_index, size)

        # Run "fused" message and aggregation (if applicable).
        if isinstance(edge_index, SparseTensor) and self.fuse and not self.__explain__:
            coll_dict = self.__collect__(
                self.__fused_user_args__, edge_index, size, kwargs
            )

            msg_aggr_kwargs = self.inspector.distribute(
                "message_and_aggregate", coll_dict
            )
            out = self.message_and_aggregate(edge_index, **msg_aggr_kwargs)

            update_kwargs = self.inspector.distribute("update", coll_dict)
            return self.update(out, **update_kwargs)

        # Otherwise, run both functions in separation.
        elif isinstance(edge_index, Tensor) or not self.fuse:
            coll_dict = self.__collect__(self.__user_args__, edge_index, size, kwargs)

            msg_kwargs = self.inspector.distribute("message", coll_dict)
            out = self.message(**msg_kwargs)

            # For `GNNExplainer`, we require a separate message and aggregate
            # procedure since this allows us to inject the `edge_mask` into the
            # message passing computation scheme.
            if self.__explain__:
                edge_mask = self.__edge_mask__.sigmoid()
                # Some ops add self-loops to `edge_index`. We need to do the
                # same for `edge_mask` (but do not train those).
                if out.size(self.node_dim) != edge_mask.size(0):
                    loop = edge_mask.new_ones(size[0])
                    edge_mask = torch.cat([edge_mask, loop], dim=0)
                assert out.size(self.node_dim) == edge_mask.size(0)
                out = out * edge_mask.view([-1] + [1] * (out.dim() - 1))

            aggr_kwargs = self.inspector.distribute("aggregate", coll_dict)
            out = self.aggregate(out, **aggr_kwargs)

            update_kwargs = self.inspector.distribute("update", coll_dict)
            return self.update(out, **update_kwargs)

    def message(self, x_j: Tensor) -> Tensor:
        r"""Constructs messages from node :math:`j` to node :math:`i`
        in analogy to :math:`\phi_{\mathbf{\Theta}}` for each edge in
        :obj:`edge_index`.
        This function can take any argument as input which was initially
        passed to :meth:`propagate`.
        Furthermore, tensors passed to :meth:`propagate` can be mapped to the
        respective nodes :math:`i` and :math:`j` by appending :obj:`_i` or
        :obj:`_j` to the variable name, *.e.g.* :obj:`x_i` and :obj:`x_j`.
        """
        return x_j

    def aggregate(
        self,
        inputs: Tensor,
        index: Tensor,
        ptr: Optional[Tensor] = None,
        dim_size: Optional[int] = None,
    ) -> Tensor:
        r"""Aggregates messages from neighbors as
        :math:`\square_{j \in \mathcal{N}(i)}`.

        Takes in the output of message computation as first argument and any
        argument which was initially passed to :meth:`propagate`.

        By default, this function will delegate its call to scatter functions
        that support "add", "mean" and "max" operations as specified in
        :meth:`__init__` by the :obj:`aggr` argument.
        """
        if ptr is not None:
            ptr = expand_left(ptr, dim=self.node_dim, dims=inputs.dim())
            return segment_csr(inputs, ptr, reduce=self.aggr)
        else:
            return scatter(
                inputs, index, dim=self.node_dim, dim_size=dim_size, reduce=self.aggr
            )

    def message_and_aggregate(self, adj_t: SparseTensor) -> Tensor:
        r"""Fuses computations of :func:`message` and :func:`aggregate` into a
        single function.
        If applicable, this saves both time and memory since messages do not
        explicitly need to be materialized.
        This function will only gets called in case it is implemented and
        propagation takes place based on a :obj:`torch_sparse.SparseTensor`.
        """
        raise NotImplementedError

    def update(self, inputs: Tensor) -> Tensor:
        r"""Updates node embeddings in analogy to
        :math:`\gamma_{\mathbf{\Theta}}` for each node
        :math:`i \in \mathcal{V}`.
        Takes in the output of aggregation as first argument and any argument
        which was initially passed to :meth:`propagate`.
        """
        return inputs

    @torch.jit.unused
    def jittable(self, typing: Optional[str] = None):
        r"""Analyzes the :class:`MessagePassing` instance and produces a new
        jittable module.

        Args:
            typing (string, optional): If given, will generate a concrete
                instance with :meth:`forward` types based on :obj:`typing`,
                *e.g.*: :obj:`"(Tensor, Optional[Tensor]) -> Tensor"`.
        """
        # Find and parse `propagate()` types to format `{arg1: type1, ...}`.
        if hasattr(self, "propagate_type"):
            prop_types = {k: sanitize(str(v)) for k, v in self.propagate_type.items()}
        else:
            source = inspect.getsource(self.__class__)
            match = re.search(r"#\s*propagate_type:\s*\((.*)\)", source)
            if match is None:
                raise TypeError(
                    "TorchScript support requires the definition of the types "
                    "passed to `propagate()`. Please specificy them via\n\n"
                    'propagate_type = {"arg1": type1, "arg2": type2, ... }\n\n'
                    "or via\n\n"
                    "# propagate_type: (arg1: type1, arg2: type2, ...)\n\n"
                    "inside the `MessagePassing` module."
                )
            prop_types = split_types_repr(match.group(1))
            prop_types = dict([re.split(r"\s*:\s*", t) for t in prop_types])

        # Parse `__collect__()` types to format `{arg:1, type1, ...}`.
        collect_types = self.inspector.types(["message", "aggregate", "update"])

        # Collect `forward()` header, body and @overload types.
        forward_types = parse_types(self.forward)
        forward_types = [resolve_types(*types) for types in forward_types]
        forward_types = list(chain.from_iterable(forward_types))

        keep_annotation = len(forward_types) < 2
        forward_header = func_header_repr(self.forward, keep_annotation)
        forward_body = func_body_repr(self.forward, keep_annotation)

        if keep_annotation:
            forward_types = []
        elif typing is not None:
            forward_types = []
            forward_body = 8 * " " + f"# type: {typing}\n{forward_body}"

        root = os.path.dirname(osp.realpath(__file__))
        with open(osp.join(root, "message_passing.jinja"), "r") as f:
            template = Template(f.read())

        uid = uuid1().hex[:6]
        cls_name = f"{self.__class__.__name__}Jittable_{uid}"
        jit_module_repr = template.render(
            uid=uid,
            module=str(self.__class__.__module__),
            cls_name=cls_name,
            parent_cls_name=self.__class__.__name__,
            prop_types=prop_types,
            collect_types=collect_types,
            user_args=self.__user_args__,
            forward_header=forward_header,
            forward_types=forward_types,
            forward_body=forward_body,
            msg_args=self.inspector.keys(["message"]),
            aggr_args=self.inspector.keys(["aggregate"]),
            msg_and_aggr_args=self.inspector.keys(["message_and_aggregate"]),
            update_args=self.inspector.keys(["update"]),
            check_input=inspect.getsource(self.__check_input__)[:-1],
            lift=inspect.getsource(self.__lift__)[:-1],
        )

        # Instantiate a class from the rendered JIT module representation.
        cls = class_from_module_repr(cls_name, jit_module_repr)
        module = cls.__new__(cls)
        module.__dict__ = self.__dict__.copy()
        module.jittable = None

        return module



from torch import nn
import torch
from torch.nn import functional as F


class DartsLayerChoice(nn.Module):
    def __init__(self, layer_choice):
        super(DartsLayerChoice, self).__init__()
        self.name = layer_choice.key
        self.op_choices = nn.ModuleDict(layer_choice.named_children())
        self.alpha = nn.Parameter(torch.randn(len(self.op_choices)) * 1e-3)

    def forward(self, *args, **kwargs):
        op_results = torch.stack(
            [op(*args, **kwargs) for op in self.op_choices.values()]
        )
        alpha_shape = [-1] + [1] * (len(op_results.size()) - 1)
        return torch.sum(op_results * F.softmax(self.alpha, -1).view(*alpha_shape), 0)

    def parameters(self):
        for _, p in self.named_parameters():
            yield p

    def named_parameters(self):
        for name, p in super(DartsLayerChoice, self).named_parameters():
            if name == "alpha":
                continue
            yield name, p

    def export(self):
        return torch.argmax(self.alpha).item()


class DartsInputChoice(nn.Module):
    def __init__(self, input_choice):
        super(DartsInputChoice, self).__init__()
        self.name = input_choice.key
        self.alpha = nn.Parameter(torch.randn(input_choice.n_candidates) * 1e-3)
        self.n_chosen = input_choice.n_chosen or 1

    def forward(self, inputs):
        inputs = torch.stack(inputs)
        alpha_shape = [-1] + [1] * (len(inputs.size()) - 1)
        return torch.sum(inputs * F.softmax(self.alpha, -1).view(*alpha_shape), 0)

    def parameters(self):
        for _, p in self.named_parameters():
            yield p

    def named_parameters(self):
        for name, p in super(DartsInputChoice, self).named_parameters():
            if name == "alpha":
                continue
            yield name, p

    def export(self):
        return torch.argsort(-self.alpha).cpu().numpy().tolist()[: self.n_chosen]



# codes in this file are reproduced from https://github.com/GraphNAS/GraphNAS with some changes.
import typing as _typ
import torch
import torch.nn.functional as F
from torch import nn
from .nni import mutables
from .base import BaseSpace
from .operation import act_map, gnn_map, act_map_nn, map_nn
from autogllight.utils.backend import BackendOperator as BK

GRAPHNAS_DEFAULT_GNN_OPS = [
    "gat_8",  # GAT with 8 heads
    "gat_6",  # GAT with 6 heads
    "gat_4",  # GAT with 4 heads
    "gat_2",  # GAT with 2 heads
    "gat_1",  # GAT with 1 heads
    "gcn",  # GCN
    "cheb",  # chebnet
    "sage",  # sage
    "arma",
    "sg",  # simplifying gcn
    "linear",  # skip connection
    "zero",  # skip connection
]

GRAPHNAS_DEFAULT_ACT_OPS = [
    # "sigmoid", "tanh", "relu", "linear",
    #  "softplus", "leaky_relu", "relu6", "elu"
    "sigmoid",
    "tanh",
    "relu",
    "linear",
    "elu",
]

GRAPHNAS_DEFAULT_CON_OPS = ["add", "product", "concat"]
# GRAPHNAS_DEFAULT_CON_OPS=[ "concat"] # for darts


class GraphNasNodeClassificationSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.9,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        gnn_ops: _typ.Sequence[_typ.Union[str, _typ.Any]] = GRAPHNAS_DEFAULT_GNN_OPS,
        act_ops: _typ.Sequence[_typ.Union[str, _typ.Any]] = GRAPHNAS_DEFAULT_ACT_OPS,
        con_ops: _typ.Sequence[_typ.Union[str, _typ.Any]] = GRAPHNAS_DEFAULT_CON_OPS,
    ):
        super().__init__()
        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.gnn_ops = gnn_ops
        self.act_ops = act_ops
        self.con_ops = con_ops
        self.dropout = dropout

    def build_graph(self):
        self.preproc0 = nn.Linear(self.input_dim, self.hidden_dim)
        self.preproc1 = nn.Linear(self.input_dim, self.hidden_dim)
        node_labels = [mutables.InputChoice.NO_KEY, mutables.InputChoice.NO_KEY]
        for layer in range(2, self.layer_number + 2):
            # input choice
            in_key = f"in_{layer}"
            self.setInputChoice(
                layer, choose_from=node_labels, n_chosen=1, key=in_key,
            )

            # operation choice
            op_key = f"op_{layer}"
            op_candidates = [
                gnn_map(op, self.hidden_dim, self.hidden_dim) for op in self.gnn_ops
            ]
            self.setLayerChoice(layer, op_candidates, key=op_key)

            # input choice candidates
            node_labels.append(op_key)

        # activation choice
        self.setLayerChoice(2 * layer, [act_map_nn(a) for a in self.act_ops], key="act")

        # for DARTS, len(con_ops) can only <=1, for dimension problems
        if len(self.con_ops) > 1:
            self.setLayerChoice(2 * layer + 1, map_nn(self.con_ops), key="concat")

        self.classifier1 = nn.Linear(
            self.hidden_dim * self.layer_number, self.output_dim
        )
        self.classifier2 = nn.Linear(self.hidden_dim, self.output_dim)

    def forward(self, data):
        # x, edges = data.x, data.edge_index  # x [2708,1433] ,[2, 10556]
        x = BK.feat(data)

        x = F.dropout(x, p=self.dropout, training=self.training)
        pprev_, prev_ = self.preproc0(x), self.preproc1(x)
        prev_nodes_out = [pprev_, prev_]
        for layer in range(2, self.layer_number + 2):
            node_in = getattr(self, f"in_{layer}")(prev_nodes_out)
            op = getattr(self, f"op_{layer}")
            node_out = BK.gconv(op, data, node_in)
            prev_nodes_out.append(node_out)
        act = getattr(self, "act")
        if len(self.con_ops) > 1:
            con = getattr(self, "concat")()
        elif len(self.con_ops) == 1:
            con = self.con_ops[0]
        else:
            con = "concat"

        states = prev_nodes_out
        if con == "concat":
            x = torch.cat(states[2:], dim=1)
        else:
            tmp = states[2]
            for i in range(3, len(states)):
                if con == "add":
                    tmp = torch.add(tmp, states[i])
                elif con == "product":
                    tmp = torch.mul(tmp, states[i])
            x = tmp
        x = act(x)
        if con == "concat":
            x = self.classifier1(x)
        else:
            x = self.classifier2(x)
        return F.log_softmax(x, dim=1)



import torch_geometric
import torch
from typing import Union, Tuple, Optional
from torch_geometric.typing import (OptPairTensor, Adj, Size, NoneType,
                                    OptTensor)
from torch import Tensor
from torch_geometric.utils import remove_self_loops, add_self_loops, softmax
import torch.nn.functional as F
from torch_sparse import SparseTensor, set_diag
from torch_geometric.nn import SAGEConv, GATConv, JumpingKnowledge
from torch_geometric.nn import GCNConv, GINConv,GraphConv,LEConv,SGConv,DenseSAGEConv,DenseGCNConv,DenseGINConv,DenseGraphConv
from .pyg_gnn_layer import GeoLayer



class GAT_mix(GATConv):
    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj,
                size: Size = None, return_attention_weights=None, edge_weight: OptTensor = None):
        r"""

        Args:
            return_attention_weights (bool, optional): If set to :obj:`True`,
                will additionally return the tuple
                :obj:`(edge_index, attention_weights)`, holding the computed
                attention weights for each edge. (default: :obj:`None`)
        """
        H, C = self.heads, self.out_channels

        x_l: OptTensor = None
        x_r: OptTensor = None
        alpha_l: OptTensor = None
        alpha_r: OptTensor = None
        if isinstance(x, Tensor):
            assert x.dim() == 2, 'Static graphs not supported in `GATConv`.'
            x_l = x_r = self.lin_l(x).view(-1, H, C)
            alpha_l = alpha_r = (x_l * self.att_l).sum(dim=-1)
        else:
            x_l, x_r = x[0], x[1]
            assert x[0].dim() == 2, 'Static graphs not supported in `GATConv`.'
            x_l = self.lin_l(x_l).view(-1, H, C)
            alpha_l = (x_l * self.att_l).sum(dim=-1)
            if x_r is not None:
                x_r = self.lin_r(x_r).view(-1, H, C)
                alpha_r = (x_r * self.att_r).sum(dim=-1)

        assert x_l is not None
        assert alpha_l is not None
        self.add_self_loops = False
        if self.add_self_loops:
            if isinstance(edge_index, Tensor):
                num_nodes = x_l.size(0)
                num_nodes = size[1] if size is not None else num_nodes
                num_nodes = x_r.size(0) if x_r is not None else num_nodes
                edge_index, _ = remove_self_loops(edge_index)
                edge_index, _ = add_self_loops(edge_index, num_nodes=num_nodes)
            elif isinstance(edge_index, SparseTensor):
                edge_index = set_diag(edge_index)

        # propagate_type: (x: OptPairTensor, alpha: OptPairTensor)
        out = self.propagate(edge_index, x=(x_l, x_r),
                             alpha=(alpha_l, alpha_r), size=size, edge_weight=edge_weight)

        alpha = self._alpha
        self._alpha = None

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if self.bias is not None:
            out += self.bias

        if isinstance(return_attention_weights, bool):
            assert alpha is not None
            if isinstance(edge_index, Tensor):
                return out, (edge_index, alpha)
            elif isinstance(edge_index, SparseTensor):
                return out, edge_index.set_value(alpha, layout='coo')
        else:
            return out

    def message(self, x_j: Tensor, alpha_j: Tensor, edge_weight: Tensor, alpha_i: OptTensor,
                index: Tensor, ptr: OptTensor, size_i: Optional[int]) -> Tensor:
        alpha = alpha_j if alpha_i is None else alpha_j + alpha_i
        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, index, ptr, size_i)
        self._alpha = alpha
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        x1 = x_j * alpha.unsqueeze(-1)
        if edge_weight is None:
            return x1
        else:
            x2 = (x1.view(-1, self.heads * self.out_channels).t() * edge_weight).t().view(-1, self.heads, self.out_channels)
            return x2

class SAGE_mix(SAGEConv):
    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj,
                size: Size = None, edge_weight: OptTensor = None) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor)
        out = self.propagate(edge_index, x=x, size=size, edge_weight=edge_weight)
        out = self.lin_l(out)

        x_r = x[1]
        if x_r is not None:
            out += self.lin_r(x_r)

        if self.normalize:
            out = F.normalize(out, p=2., dim=-1)

        return out

    def message(self, x_j: Tensor, edge_weight: Tensor) -> Tensor:
        return x_j if edge_weight is None else edge_weight.view(-1, 1) * x_j

class GIN_mix(GINConv):
    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj,
                size: Size = None, edge_weight: OptTensor = None) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor)
        out = self.propagate(edge_index, x=x, size=size, edge_weight=edge_weight)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    def message(self, x_j: Tensor, edge_weight: Tensor) -> Tensor:
        return x_j if edge_weight is None else edge_weight.view(-1, 1) * x_j

class Geolayer_mix(GeoLayer):
    def forward(self, x, edge_index, size=None, edge_weight: OptTensor = None):
        """"""
        # rm self loop
        # if size is None and torch.is_tensor(x):
        #     edge_index, _ = remove_self_loops(edge_index)
        #     edge_index, _ = add_self_loops(edge_index, num_nodes=x.size(0))

        # prepare
        if torch.is_tensor(x):
            x = torch.mm(x, self.weight).view(-1, self.heads, self.out_channels)
        else:
            x = (None if x[0] is None else torch.matmul(x[0], self.weight).view(-1, self.heads, self.out_channels),
                 None if x[1] is None else torch.matmul(x[1], self.weight).view(-1, self.heads, self.out_channels))
        num_nodes = x.size(0) if torch.is_tensor(x) else size[0]
        return self.propagate(edge_index, size=size, x=x, num_nodes=num_nodes, edge_weight=edge_weight)
    def message(self, x_i, x_j, edge_index, num_nodes, edge_weight=None):
        if self.att_type == "const":
            if self.training and self.dropout > 0:
                x_j = F.dropout(x_j, p=self.dropout, training=True)
            neighbor = x_j

        else:
            # Compute attention coefficients.
            alpha = self.apply_attention(edge_index, num_nodes, x_i, x_j)
            alpha = softmax(alpha, edge_index[0], ptr=None, num_nodes=num_nodes)
            # Sample attention coefficients stochastically.
            if self.training and self.dropout > 0:
                alpha = F.dropout(alpha, p=self.dropout, training=True)

            neighbor = x_j * alpha.view(-1, self.heads, 1)
            # if self.pool_dim > 0:
            #    for layer in self.pool_layer:
            #        neighbor = layer(neighbor)
        if edge_weight is None:
            return neighbor
        else:
            x2 = (neighbor.view(-1, self.heads * self.out_channels).t() * edge_weight).t().view(-1, self.heads, self.out_channels)
            return x2











from .operations import *

OPS = {
    "ZERO": lambda indim, outdim, dropout, concat=False: Zero(indim, outdim),
    "IDEN": lambda indim, outdim, dropout, concat=False: Identity(),
    "GCN": lambda indim, outdim, dropout, concat=False: GCNConv(
        indim, outdim, add_self_loops=False
    ),
    "SAGE-MEAN": lambda indim, outdim, dropout, concat=False: SAGEConv(indim, outdim),
    "GAT16": lambda indim, outdim, dropout, concat=False: GATConv(
        indim, outdim, dropout=dropout, heads=16, concat=False, add_self_loops=False
    )
    if not concat
    else GATConv(
        indim,
        outdim // 16,
        dropout=dropout,
        heads=16,
        concat=True,
        add_self_loops=False,
    ),
    "GAT2": lambda indim, outdim, dropout, concat=False: GATConv(
        indim, outdim, dropout=dropout, heads=2, concat=False, add_self_loops=False
    )
    if not concat
    else GATConv(
        indim, outdim // 2, dropout=dropout, heads=2, concat=True, add_self_loops=False
    ),
    "GAT4": lambda indim, outdim, dropout, concat=False: GATConv(
        indim, outdim, dropout=dropout, heads=4, concat=False, add_self_loops=False
    )
    if not concat
    else GATConv(
        indim, outdim // 4, dropout=dropout, heads=4, concat=True, add_self_loops=False
    ),
    "GAT8": lambda indim, outdim, dropout, concat=False: GATConv(
        indim, outdim, dropout=dropout, heads=8, concat=False, add_self_loops=False
    )
    if not concat
    else GATConv(
        indim, outdim // 8, dropout=dropout, heads=8, concat=True, add_self_loops=False
    ),
    "GAT1": lambda indim, outdim, dropout, concat=False: GATConv(
        indim, outdim, dropout=dropout, heads=1, concat=False, add_self_loops=False
    ),
    "LIN": lambda indim, outdim, dropout, concat=False: Linear(indim, outdim),
    "ARMA": lambda indim, outdim, dropout, concat=False: ARMAConv(indim, outdim),
    "CHEB": lambda indim, outdim, dropout, concat=False: ChebConv(indim, outdim, 2),
    "SGC": lambda indim, outdim, dropout, concat=False: SGConv(
        indim, outdim, add_self_loops=False
    ),
}

PRIMITIVES = list(OPS.keys())



import math
import json
import torch
import random
import torch.nn as nn
from .lr_file import PolynomialDecayLR


def init_params(module, n_layers):
    if isinstance(module, nn.Linear):
        module.weight.data.normal_(mean=0.0, std=0.02 / math.sqrt(n_layers))
        if module.bias is not None:
            module.bias.data.zero_()
    if isinstance(module, nn.Embedding):
        module.weight.data.normal_(mean=0.0, std=0.02)


class GraphTransformer(nn.Module):
    def __init__(
        self,
        n_layers,
        num_heads,
        hidden_dim,
        dropout_rate,
        intput_dropout_rate,
        weight_decay,
        ffn_dim,
        dataset_name,
        warmup_updates,
        tot_updates,
        peak_lr,
        end_lr,
        edge_type,
        multi_hop_max_dist,
        attention_dropout_rate,
        num_class,
        lap_dim,
        svd_dim,
        path,
    ):
        super().__init__()

        self.num_heads = num_heads
        if dataset_name == 'PROTEINS':
            self.atom_encoder = nn.Embedding(8, hidden_dim, padding_idx=0)
            self.edge_encoder = nn.Embedding(4, num_heads, padding_idx=0)
            self.edge_type = edge_type
            self.edge_dis_encoder = nn.Embedding(
                40 * num_heads * num_heads, 1)
            self.spatial_pos_encoder = nn.Embedding(512, num_heads, padding_idx=0)
            self.in_degree_encoder = nn.Embedding(
                64, hidden_dim, padding_idx=0)
            self.out_degree_encoder = nn.Embedding(
                64, hidden_dim, padding_idx=0)
        else:
            self.atom_encoder = nn.Embedding(
                256, hidden_dim, padding_idx=0)
            self.edge_encoder = nn.Embedding(
                256, num_heads, padding_idx=0)
            self.edge_type = edge_type
            self.edge_dis_encoder = nn.Embedding(
                64 * num_heads * num_heads, 1)
            self.spatial_pos_encoder = nn.Embedding(512, num_heads, padding_idx=0)
            self.in_degree_encoder = nn.Embedding(
                128, hidden_dim, padding_idx=0)
            self.out_degree_encoder = nn.Embedding(
                128, hidden_dim, padding_idx=0)

        self.input_dropout = nn.Dropout(intput_dropout_rate)
        encoders = [EncoderLayer(hidden_dim, ffn_dim, dropout_rate, attention_dropout_rate, num_heads)
                    for _ in range(n_layers)]
        self.layers = nn.ModuleList(encoders)
        self.final_ln = nn.LayerNorm(hidden_dim)

        self.downstream_out_proj = nn.Linear(hidden_dim, num_class)
        self.pma_linear = nn.Linear(multi_hop_max_dist, num_heads)
        self.lap_linear = nn.Linear(lap_dim, hidden_dim)
        self.svd_linear = nn.Linear(svd_dim, hidden_dim)

        self.graph_token = nn.Embedding(1, hidden_dim)
        self.graph_token_virtual_distance = nn.Embedding(1, num_heads)

        self.warmup_updates = warmup_updates
        self.tot_updates = tot_updates
        self.peak_lr = peak_lr
        self.end_lr = end_lr
        self.weight_decay = weight_decay
        self.multi_hop_max_dist = multi_hop_max_dist

        self.hidden_dim = hidden_dim
        self.apply(lambda module: init_params(module, n_layers=n_layers))
        self.dataset_name = dataset_name
        self.encodings = True
        self.path = path

    def forward(self, batched_data, params=None):
        if params == None and self.encodings:
            # Initialize
            attn_bias, spatial_pos, x = batched_data.attn_bias, batched_data.spatial_pos, batched_data.x
            in_degree, out_degree = batched_data.in_degree, batched_data.in_degree
            edge_input, attn_edge_type = batched_data.edge_input, batched_data.attn_edge_type
            lap_pos_enc, svd_pos_enc, pma_att_enc = batched_data.lap_pos_enc, batched_data.svd_pos_enc, batched_data.pma_att_enc
            # Attention Matrix: Spatial Encoding, Edge Encoding, Proximity-Enhanced Multi-Head Attention
            # Positional Embedding: Centrality Encoding, Laplacian Eigenvector, SVD-based Positional Encoding
            # Mask: Mask-n

            # print("atom_encoder minimal num_embeddings: {} - {} = {}".format(x.max(), x.min(), x.max() - x.min()))
            # print("edge_encoder minimal num_embeddings: {} - {} = {}".format(edge_input.max(), edge_input.min(), edge_input.max() - edge_input.min()))
            # print("spatial_pos_encoder minimal num_embeddings: {} - {} = {}".format(spatial_pos.max(), spatial_pos.min(), spatial_pos.max() - spatial_pos.min()))
            # print("in_degree_encoder minimal num_embeddings: {} - {} = {}".format(in_degree.max(), in_degree.min(), in_degree.max() - in_degree.min()))

            # ==================== Attention Matrix Start ====================
            # graph_attn_bias
            n_graph, n_node = x.size()[:2]
            graph_attn_bias = attn_bias.clone() # attn_bias [n_graph, n_node+1, n_node+1]
            graph_attn_bias = graph_attn_bias.unsqueeze(1).repeat(1, self.num_heads, 1, 1)  # [n_graph, n_head, n_node+1, n_node+1]
            # reset spatial pos here
            t = self.graph_token_virtual_distance.weight.view(1, self.num_heads, 1)
            graph_attn_bias[:, :, 1:, 0] = graph_attn_bias[:, :, 1:, 0] + t
            graph_attn_bias[:, :, 0, :] = graph_attn_bias[:, :, 0, :] + t

            # spatial pos: Spatial Encoding
            # [n_graph, n_node, n_node, n_head] -> [n_graph, n_head, n_node, n_node]
            spatial_pos_bias = self.spatial_pos_encoder(spatial_pos).permute(0, 3, 1, 2)
            graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + spatial_pos_bias

            # edge feature: Edge Encoding
            spatial_pos_ = spatial_pos.clone()
            spatial_pos_[spatial_pos_ == 0] = 1  # set pad to 1
            # set 1 to 1, x > 1 to x - 1
            spatial_pos_ = torch.where(spatial_pos_ > 1, spatial_pos_ - 1, spatial_pos_)
            spatial_pos_ = spatial_pos_.clamp(0, self.multi_hop_max_dist)
            edge_input = edge_input[:, :, :, :self.multi_hop_max_dist, :]
            # [n_graph, n_node, n_node, max_dist, n_head]
            edge_input = self.edge_encoder(edge_input).mean(-2)
            max_dist = edge_input.size(-2)
            edge_input_flat = edge_input.permute(3, 0, 1, 2, 4).reshape(max_dist, -1, self.num_heads)
            edge_input_flat = torch.bmm(edge_input_flat, self.edge_dis_encoder.weight.reshape(
                -1, self.num_heads, self.num_heads)[:max_dist, :, :])
            edge_input = edge_input_flat.reshape(max_dist, n_graph, n_node, n_node, self.num_heads).permute(1, 2, 3, 0, 4)
            edge_input = (edge_input.sum(-2) / (spatial_pos_.float().unsqueeze(-1))).permute(0, 3, 1, 2)
            graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + edge_input
            graph_attn_bias = graph_attn_bias + attn_bias.unsqueeze(1)  # reset

            # PMA: Proximity-Enhanced Multi-Head Attention
            proximity_enhanced_bias = self.pma_linear(pma_att_enc).permute(0, 3, 1, 2)
            graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + proximity_enhanced_bias
            # ===================== Attention Matrix End =====================

            # ==================== Positional Embedding Start ====================
            # node feauture + graph token
            node_feature = self.atom_encoder(x).sum(dim=-2) # x [n_graph, n_node, feature_dim]
            # [n_graph, n_node, n_hidden]

            # Centrality Encoding
            # node_feature = node_feature + self.in_degree_encoder(in_degree) + self.out_degree_encoder(out_degree)
            positional_embedding = self.in_degree_encoder(in_degree) + self.out_degree_encoder(out_degree)

            # Laplacian Eigenvector
            lap_feature = self.lap_linear(lap_pos_enc)
            # node_feature = node_feature + lap_feature
            positional_embedding = positional_embedding + lap_feature

            # SVD-based Positional Encoding
            svd_feature = self.svd_linear(svd_pos_enc)
            # node_feature = node_feature + svd_feature
            positional_embedding = positional_embedding + svd_feature

            graph_token_feature = self.graph_token.weight.unsqueeze(0).repeat(n_graph, 1, 1)
            graph_node_feature = torch.cat([graph_token_feature, node_feature], dim=1)
            # [n_graph, n_node+1, hidden_dim]
            # ===================== Positional Embedding End =====================

            # transfomrer encoder
            output = self.input_dropout(graph_node_feature)
            for enc_layer in self.layers:
                output[:, 1:] = output[:, 1:] + positional_embedding
                output = enc_layer(output, graph_attn_bias, spatial_pos=spatial_pos)
            output = self.final_ln(output)

            # output part
            output = self.downstream_out_proj(output[:, 0, :])
            return output
        elif params == None:
            # Initialize
            attn_bias, spatial_pos, x = batched_data.attn_bias, batched_data.spatial_pos, batched_data.x
            in_degree, out_degree = batched_data.in_degree, batched_data.in_degree
            edge_input, attn_edge_type = batched_data.edge_input, batched_data.attn_edge_type
            lap_pos_enc, svd_pos_enc, pma_att_enc = batched_data.lap_pos_enc, batched_data.svd_pos_enc, batched_data.pma_att_enc
            # Attention Matrix: Spatial Encoding, Edge Encoding, Proximity-Enhanced Multi-Head Attention
            # Positional Embedding: Centrality Encoding, Laplacian Eigenvector, SVD-based Positional Encoding
            # Mask: Mask-n

            # ==================== Attention Matrix Start ====================
            # graph_attn_bias
            n_graph, n_node = x.size()[:2]
            graph_attn_bias = attn_bias.clone() # attn_bias [n_graph, n_node+1, n_node+1]
            graph_attn_bias = graph_attn_bias.unsqueeze(1).repeat(1, self.num_heads, 1, 1)  # [n_graph, n_head, n_node+1, n_node+1]
            # reset spatial pos here
            t = self.graph_token_virtual_distance.weight.view(1, self.num_heads, 1)
            graph_attn_bias[:, :, 1:, 0] = graph_attn_bias[:, :, 1:, 0] + t
            graph_attn_bias[:, :, 0, :] = graph_attn_bias[:, :, 0, :] + t

            # spatial pos: Spatial Encoding
            # [n_graph, n_node, n_node, n_head] -> [n_graph, n_head, n_node, n_node]
            spatial_pos_bias = self.spatial_pos_encoder(spatial_pos).permute(0, 3, 1, 2)
            graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + spatial_pos_bias

            # edge feature: Edge Encoding
            spatial_pos_ = spatial_pos.clone()
            spatial_pos_[spatial_pos_ == 0] = 1  # set pad to 1
            # set 1 to 1, x > 1 to x - 1
            spatial_pos_ = torch.where(spatial_pos_ > 1, spatial_pos_ - 1, spatial_pos_)
            spatial_pos_ = spatial_pos_.clamp(0, self.multi_hop_max_dist)
            edge_input = edge_input[:, :, :, :self.multi_hop_max_dist, :]
            # [n_graph, n_node, n_node, max_dist, n_head]
            edge_input = self.edge_encoder(edge_input).mean(-2)
            max_dist = edge_input.size(-2)
            edge_input_flat = edge_input.permute(3, 0, 1, 2, 4).reshape(max_dist, -1, self.num_heads)
            edge_input_flat = torch.bmm(edge_input_flat, self.edge_dis_encoder.weight.reshape(
                -1, self.num_heads, self.num_heads)[:max_dist, :, :])
            edge_input = edge_input_flat.reshape(max_dist, n_graph, n_node, n_node, self.num_heads).permute(1, 2, 3, 0, 4)
            edge_input = (edge_input.sum(-2) / (spatial_pos_.float().unsqueeze(-1))).permute(0, 3, 1, 2)
            graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + edge_input
            graph_attn_bias = graph_attn_bias + attn_bias.unsqueeze(1)  # reset
            # ===================== Attention Matrix End =====================

            # ==================== Positional Embedding Start ====================
            # node feauture + graph token
            node_feature = self.atom_encoder(x).sum(dim=-2) # x [n_graph, n_node, feature_dim]
            # [n_graph, n_node, n_hidden]

            # Centrality Encoding
            node_feature = node_feature + self.in_degree_encoder(in_degree) + self.out_degree_encoder(out_degree)

            graph_token_feature = self.graph_token.weight.unsqueeze(0).repeat(n_graph, 1, 1)
            graph_node_feature = torch.cat([graph_token_feature, node_feature], dim=1)
            # [n_graph, n_node+1, hidden_dim]
            # ===================== Positional Embedding End =====================

            # transfomrer encoder
            output = self.input_dropout(graph_node_feature)
            for enc_layer in self.layers:
                output = enc_layer(output, graph_attn_bias)
            output = self.final_ln(output)

            # output part
            output = self.downstream_out_proj(output[:, 0, :])
            return output
        else:
            # Initialize
            attn_bias, spatial_pos, x = batched_data.attn_bias, batched_data.spatial_pos, batched_data.x
            in_degree, out_degree = batched_data.in_degree, batched_data.in_degree
            edge_input, attn_edge_type = batched_data.edge_input, batched_data.attn_edge_type
            lap_pos_enc, svd_pos_enc, pma_att_enc = batched_data.lap_pos_enc, batched_data.svd_pos_enc, batched_data.pma_att_enc
            # Attention Matrix: Spatial Encoding, Edge Encoding, Proximity-Enhanced Multi-Head Attention
            # Positional Embedding: Centrality Encoding, Laplacian Eigenvector, SVD-based Positional Encoding
            # Mask: Mask-n

            # ==================== Attention Matrix Start ====================
            # graph_attn_bias
            n_graph, n_node = x.size()[:2]
            graph_attn_bias = attn_bias.clone() # attn_bias [n_graph, n_node+1, n_node+1]
            graph_attn_bias = graph_attn_bias.unsqueeze(1).repeat(1, self.num_heads, 1, 1)  # [n_graph, n_head, n_node+1, n_node+1]
            # reset spatial pos here
            t = self.graph_token_virtual_distance.weight.view(1, self.num_heads, 1)
            graph_attn_bias[:, :, 1:, 0] = graph_attn_bias[:, :, 1:, 0] + t
            graph_attn_bias[:, :, 0, :] = graph_attn_bias[:, :, 0, :] + t

            # spatial pos: Spatial Encoding
            # [n_graph, n_node, n_node, n_head] -> [n_graph, n_head, n_node, n_node]
            spatial_pos_bias = self.spatial_pos_encoder(spatial_pos).permute(0, 3, 1, 2)
            # graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + spatial_pos_bias

            # edge feature: Edge Encoding
            spatial_pos_ = spatial_pos.clone()
            spatial_pos_[spatial_pos_ == 0] = 1  # set pad to 1
            # set 1 to 1, x > 1 to x - 1
            spatial_pos_ = torch.where(spatial_pos_ > 1, spatial_pos_ - 1, spatial_pos_)
            spatial_pos_ = spatial_pos_.clamp(0, self.multi_hop_max_dist)
            edge_input = edge_input[:, :, :, :self.multi_hop_max_dist, :]
            # [n_graph, n_node, n_node, max_dist, n_head]
            edge_input = self.edge_encoder(edge_input).mean(-2)
            max_dist = edge_input.size(-2)
            edge_input_flat = edge_input.permute(3, 0, 1, 2, 4).reshape(max_dist, -1, self.num_heads)
            edge_input_flat = torch.bmm(edge_input_flat, self.edge_dis_encoder.weight.reshape(
                -1, self.num_heads, self.num_heads)[:max_dist, :, :])
            edge_input = edge_input_flat.reshape(max_dist, n_graph, n_node, n_node, self.num_heads).permute(1, 2, 3, 0, 4)
            edge_input = (edge_input.sum(-2) / (spatial_pos_.float().unsqueeze(-1))).permute(0, 3, 1, 2)
            # graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + edge_input
            graph_attn_bias = graph_attn_bias + attn_bias.unsqueeze(1)  # reset

            # PMA: Proximity-Enhanced Multi-Head Attention
            proximity_enhanced_bias = self.pma_linear(pma_att_enc).permute(0, 3, 1, 2)
            # graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[:, :, 1:, 1:] + proximity_enhanced_bias
            # ===================== Attention Matrix End =====================

            # ==================== Positional Embedding Start ====================
            # node feauture + graph token
            node_feature = self.atom_encoder(x).sum(dim=-2) # x [n_graph, n_node, feature_dim]
            # [n_graph, n_node, n_hidden]

            # Centrality Encoding
            cen_feature = self.in_degree_encoder(in_degree) + self.out_degree_encoder(out_degree)

            # Laplacian Eigenvector
            lap_feature = self.lap_linear(lap_pos_enc)

            # SVD-based Positional Encoding
            svd_feature = self.svd_linear(svd_pos_enc)

            graph_token_feature = self.graph_token.weight.unsqueeze(0).repeat(n_graph, 1, 1)
            graph_node_feature = torch.cat([graph_token_feature, node_feature], dim=1)
            # [n_graph, n_node+1, hidden_dim]
            # ===================== Positional Embedding End =====================

            # transfomrer encoder
            output = self.input_dropout(graph_node_feature)
            layers = self.layers[:params[0]]
            for i, enc_layer in enumerate(layers):
                # 这里我们用人话来解释一下生成的 params 都是啥，还有它们是怎么工作的
                # 本函数的返回值是 params = [depth, layers] 其中 depth 为层数， layers 为参数
                # layers = [layer_1, layer_2, ..., layers_depth] 其中 layer_n 为每层的参数
                # layer_n = [shape, PE, AT, mask] 其中 shape 为 Transformer 的架构超参数， PE 与 AT 为 encoding 相关参数， mask 为 mask 相关参数
                # shape = [hidden_in, num_heads, att_size, hidden_mid, ffn_size] 其中 num_heads 为头数，其他均为维度数
                # PE = [cen, eig, svd] 其中三者均为布尔类型变量，表征这一编码是否选取
                # AT = [spa, edg, pma] 其中三者均为布尔类型变量，表征这一编码是否选取
                # mask 表征是否施加，以及施加何种程度的 mask ， 0 表示不施加 mask ，正值则表示 mask 掉所有 hop 数高于 mask 值的边
                layer_i = params[1][i]
                node_feature = 0
                node_feature = node_feature + layer_i[1][0] * cen_feature
                node_feature = node_feature + layer_i[1][1] * lap_feature
                node_feature = node_feature + layer_i[1][2] * svd_feature
                output[:, 1:] = output[:, 1:] + node_feature
                graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[
                    :, :, 1:, 1:] + layer_i[2][0] * spatial_pos_bias
                graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[
                    :, :, 1:, 1:] + layer_i[2][1] * edge_input
                graph_attn_bias[:, :, 1:, 1:] = graph_attn_bias[
                    :, :, 1:, 1:] + layer_i[2][2] * proximity_enhanced_bias
                output = enc_layer(output, graph_attn_bias, layer_i[0], spatial_pos)
            output = self.final_ln(output)

            # output part
            output = self.downstream_out_proj(output[:, 0, :])
            return output

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(
            self.parameters(), lr=self.peak_lr, weight_decay=self.weight_decay)
        lr_scheduler = {
            'scheduler': PolynomialDecayLR(
                optimizer,
                warmup_updates=self.warmup_updates,
                tot_updates=self.tot_updates,
                lr=self.peak_lr,
                end_lr=self.end_lr,
                power=1.0,
            ),
            'name': 'learning_rate',
            'interval': 'step',
            'frequency': 1,
        }
        return optimizer, lr_scheduler

    def gen_params(self):
        # 这里我们用人话来解释一下生成的 params 都是啥，还有它们是怎么工作的
        # 本函数的返回值是 params = [depth, layers] 其中 depth 为层数， layers 为参数
        # layers = [layer_1, layer_2, ..., layers_depth] 其中 layer_n 为每层的参数
        # layer_n = [shape, PE, AT, mask] 其中 shape 为 Transformer 的架构超参数， PE 与 AT 为 encoding 相关参数， mask 为 mask 相关参数
        # shape = [hidden_in, num_heads, att_size, hidden_mid, ffn_size] 其中 num_heads 为头数，其他均为维度数
        # PE = [cen, eig, svd] 其中三者均为布尔类型变量，表征这一编码是否选取
        # AT = [spa, edg, pma] 其中三者均为布尔类型变量，表征这一编码是否选取
        # mask 表征是否施加，以及施加何种程度的 mask ， 0 表示不施加 mask ，正值则表示 mask 掉所有 hop 数高于 mask 值的边
        # 经研究，于2022年5月11日将 mask 移入 shape 的末尾
        assert(self.path != '')
        with open(self.path, 'r') as f:
            dic = json.load(f)
        depth = random.choice(dic['depth'])
        layers = []
        for _ in range(0, depth):
            layer = []
            hidden_in = random.choice(dic['hidden_in'])
            num_heads = random.choice(dic['num_heads'])
            att_size = random.choice(dic['att_size'])
            hidden_mid = random.choice(dic['hidden_mid'])
            ffn_size = random.choice(dic['ffn_size'])
            mask = random.choice(dic['mask'])
            layer.append((hidden_in, num_heads, att_size, hidden_mid, ffn_size, mask))
            cen = random.choice([True, False])
            eig = random.choice([True, False])
            svd = random.choice([True, False])
            layer.append((cen, eig, svd))
            spa = random.choice([True, False])
            edg = random.choice([True, False])
            pma = random.choice([True, False])
            layer.append((spa, edg, pma))
            layers.append(tuple(layer))
        return (depth, tuple(layers))

    def to_graphormer(self):
        self.encodings = False


class FeedForwardNetwork(nn.Module):
    def __init__(self, hidden_size, ffn_size, dropout_rate):
        super(FeedForwardNetwork, self).__init__()

        self.layer1 = nn.Linear(hidden_size, ffn_size)
        self.gelu = nn.GELU()
        self.layer2 = nn.Linear(ffn_size, hidden_size)

    def forward(self, x, shape=None):
        if shape == None:
            x = self.layer1(x)
            x = self.gelu(x)
            x = self.layer2(x)
        else:
            x[:, :, shape[3]:] = 0
            x = self.layer1(x)
            x[:, :, shape[4]:] = 0
            x = self.gelu(x)
            x = self.layer2(x)
        return x


class MultiHeadAttention(nn.Module):
    def __init__(self, hidden_size, attention_dropout_rate, num_heads):
        super(MultiHeadAttention, self).__init__()

        self.num_heads = num_heads

        self.att_size = att_size = hidden_size // num_heads
        self.scale = att_size ** -0.5

        self.linear_q = nn.Linear(hidden_size, num_heads * att_size)
        self.linear_k = nn.Linear(hidden_size, num_heads * att_size)
        self.linear_v = nn.Linear(hidden_size, num_heads * att_size)
        self.att_dropout = nn.Dropout(attention_dropout_rate)

        self.output_layer = nn.Linear(num_heads * att_size, hidden_size)

    def forward(self, q, k, v, attn_bias=None, shape=None, spatial_pos=None):
        orig_q_size = q.size()

        d_k = self.att_size
        d_v = self.att_size
        batch_size = q.size(0)

        if shape != None:
            q[:, :, shape[0]:] = 0
            k[:, :, shape[0]:] = 0
            v[:, :, shape[0]:] = 0

        # head_i = Attention(Q(W^Q)_i, K(W^K)_i, V(W^V)_i)
        q = self.linear_q(q).view(batch_size, -1, self.num_heads, d_k)
        k = self.linear_k(k).view(batch_size, -1, self.num_heads, d_k)
        v = self.linear_v(v).view(batch_size, -1, self.num_heads, d_v)

        if shape != None:
            q[:, :, shape[1]:, shape[2]:] = 0
            k[:, :, shape[1]:, shape[2]:] = 0
            v[:, :, shape[1]:, shape[2]:] = 0

        q = q.transpose(1, 2)                  # [b, h, q_len, d_k]
        v = v.transpose(1, 2)                  # [b, h, v_len, d_v]
        k = k.transpose(1, 2).transpose(2, 3)  # [b, h, d_k, k_len]

        # Scaled Dot-Product Attention.
        # Attention(Q, K, V) = softmax((QK^T)/sqrt(d_k))V
        if shape != None:
            q = q * (shape[2] ** -0.5)
        else:
            q = q * self.scale
        x = torch.matmul(q, k)  # [b, h, q_len, k_len]
        if attn_bias is not None:
            x = x + attn_bias
        if shape is not None and spatial_pos is not None:
            spatial_pos_ = spatial_pos.clone()
            spatial_pos_[spatial_pos_ > shape[5]] = -1
            spatial_pos_[spatial_pos_ != -1] = 0
            spatial_pos_[spatial_pos_ != 0] = -1e8
            x = x.permute(0, 2, 3, 1)
            x[:, 1:, 1:] = x[:, 1:, 1:] + spatial_pos_.unsqueeze(-1)
            x = x.permute(0, 3, 1, 2)

        x = torch.softmax(x, dim=3)
        x = self.att_dropout(x)
        x = x.matmul(v)  # [b, h, q_len, attn]

        x = x.transpose(1, 2).contiguous()  # [b, q_len, h, attn]
        x = x.view(batch_size, -1, self.num_heads * d_v)

        x = self.output_layer(x)

        assert x.size() == orig_q_size
        return x


class EncoderLayer(nn.Module):
    def __init__(self, hidden_size, ffn_size, dropout_rate, attention_dropout_rate, num_heads):
        super(EncoderLayer, self).__init__()

        self.self_attention_norm = nn.LayerNorm(hidden_size)
        self.self_attention = MultiHeadAttention(
            hidden_size, attention_dropout_rate, num_heads)
        self.self_attention_dropout = nn.Dropout(dropout_rate)

        self.ffn_norm = nn.LayerNorm(hidden_size)
        self.ffn = FeedForwardNetwork(hidden_size, ffn_size, dropout_rate)
        self.ffn_dropout = nn.Dropout(dropout_rate)

    def forward(self, x, attn_bias=None, shape=None, spatial_pos=None):
        y = self.self_attention_norm(x)
        y = self.self_attention(y, y, y, attn_bias, shape, spatial_pos)
        y = self.self_attention_dropout(y)
        x = x + y

        y = self.ffn_norm(x)
        y = self.ffn(y, shape)
        y = self.ffn_dropout(y)
        x = x + y
        return x



# "Graph differentiable architecture search with structure optimization" NeurIPS 21'

import logging
import torch
import torch.optim
from .base import BaseNAS
from ..estimator.base import BaseEstimator
from ..space import BaseSpace

from torch.autograd import Variable
import time
import torch.optim as optim
from tqdm import tqdm, trange

_logger = logging.getLogger(__name__)


class Gasso(BaseNAS):
    """
    GASSO trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    warmup_epochs : int
        Number of epochs planned for warming up.
    workers : int
        Workers for data loading.
    model_lr : float
        Learning rate to optimize the model.
    model_wd : float
        Weight decay to optimize the model.
    arch_lr : float
        Learning rate to optimize the architecture.
    stru_lr : float
        Learning rate to optimize the structure.
    lamb : float
        The parameter to control the influence of hidden feature smoothness
    device : str or torch.device
        The device of the whole process
    """

    def __init__(
        self,
        num_epochs=250,
        warmup_epochs=10,
        model_lr=0.01,
        model_wd=1e-4,
        arch_lr=0.03,
        stru_lr=0.04,
        lamb=0.6,
        device="auto",
        disable_progress=False,
    ):
        super().__init__(device=device)
        self.num_epochs = num_epochs
        self.warmup_epochs = warmup_epochs
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.arch_lr = arch_lr
        self.stru_lr = stru_lr
        self.lamb = lamb
        self.disable_progress = disable_progress

    def train_stru(self, model, optimizer, data):
        # forward
        model.train()
        # data[0].adj = self.adjs
        logits = model(data[0].to(self.device), self.adjs).detach()
        loss = 0
        for adj in self.adjs:
            e1 = adj[0][0]
            e2 = adj[0][1]
            ew = adj[1]
            diff = (logits[e1] - logits[e2]).pow(2).sum(1)
            smooth = (diff * torch.sigmoid(ew)).sum()
            dist = (ew * ew).sum()
            loss += self.lamb * smooth + dist

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        train_loss = loss.item()
        del logits

    def _infer(self, model: BaseSpace, dataset, estimator: BaseEstimator, mask="train"):
        # dataset[0].adj = self.adjs
        metric, loss = estimator.infer(model, dataset, mask=mask, adjs=self.adjs)
        return metric, loss

    def prepare(self, dset):
        """Train Pro-GNN.
        """
        data = dset[0]
        self.ews = []
        self.edges = data.edge_index.to(self.device)
        edge_weight = torch.ones(self.edges.size(1)).to(self.device)

        self.adjs = []
        for i in range(self.steps):
            edge_weight = Variable(edge_weight * 1.0, requires_grad=True).to(
                self.device
            )
            self.ews.append(edge_weight)
            self.adjs.append((self.edges, edge_weight))

    def fit(self, data):
        self.optimizer = optim.Adam(
            self.space.parameters(), lr=self.model_lr, weight_decay=self.model_wd
        )
        self.arch_optimizer = optim.Adam(
            self.space.arch_parameters(), lr=self.arch_lr, betas=(0.5, 0.999)
        )
        self.stru_optimizer = optim.SGD(self.ews, lr=self.stru_lr)

        # Train model
        best_performance = 0
        min_val_loss = float("inf")
        min_train_loss = float("inf")

        t_total = time.time()
        with trange(self.num_epochs, disable=self.disable_progress) as bar:
            for epoch in bar:
                self.space.train()
                self.optimizer.zero_grad()
                _, loss = self._infer(self.space, data, self.estimator, "train")
                loss.backward()
                self.optimizer.step()

                if epoch < 20:
                    continue
                self.train_stru(self.space, self.stru_optimizer, data)

                self.arch_optimizer.zero_grad()
                _, loss = self._infer(self.space, data, self.estimator, "train")
                loss.backward()
                self.arch_optimizer.step()

                self.space.eval()
                train_acc, _ = self._infer(self.space, data, self.estimator, "train")
                val_acc, val_loss = self._infer(self.space, data, self.estimator, "val")
                if val_loss < min_val_loss:
                    min_val_loss = val_loss
                    best_performance = val_acc
                    self.space.keep_prediction()

                bar.set_postfix(train_acc=train_acc["acc"], val_acc=val_acc["acc"])
                # print("acc:" + str(train_acc) + " val_acc" + str(val_acc))

        return best_performance, min_val_loss

    def search(self, space: BaseSpace, dataset, estimator):
        self.estimator = estimator
        self.space = space.to(self.device)
        self.steps = space.steps
        self.prepare(dataset)
        perf, val_loss = self.fit(dataset)
        return space.parse_model(None)



import importlib
import os
from .base import BaseHPOptimizer, RangeHP, LogRangeHP, ChoiceHP
from .auto_module import AutoModule

HPO_DICT = {}


def register_hpo(name):
    def register_hpo_cls(cls):
        if name in HPO_DICT:
            raise ValueError("Cannot register duplicate HP Optimizer ({})".format(name))
        if not issubclass(cls, BaseHPOptimizer):
            raise ValueError(
                "Model ({}: {}) must extend HPOptimizer".format(name, cls.__name__)
            )
        HPO_DICT[name] = cls
        return cls

    return register_hpo_cls


from .anneal_advisorhpo import AnnealAdvisorHPO
from .autone import AutoNE
from .bayes_advisor import BayesAdvisor
from .cmaes_advisorchoco import CmaesAdvisorChoco
from .grid_advisor import GridAdvisor
from .mocmaes_advisorchoco import MocmaesAdvisorChoco
from .quasi_advisorchoco import QuasiAdvisorChoco
from .rand_advisor import RandAdvisor
from .tpe_advisorhpo import TpeAdvisorHPO


def build_hpo_from_name(name: str, *args, **kwargs) -> BaseHPOptimizer:
    """
    Parameters
    ----------
    name: ``str``
        the name of hpo module.

    Returns
    -------
    BaseHPOptimizer:
        the HPO built using default parameters

    Raises
    ------
    AssertionError
        If an invalid name is passed in
    """
    assert name in HPO_DICT, "HPO module do not have name " + name
    return HPO_DICT[name](*args, **kwargs)


__all__ = [
    "AutoModule",
    "BaseHPOptimizer",
    "AnnealAdvisorHPO",
    "AutoNE",
    "BayesAdvisor",
    "CmaesAdvisorChoco",
    "GridAdvisor",
    "MocmaesAdvisorChoco",
    "QuasiAdvisorChoco",
    "RandAdvisor",
    "TpeAdvisorHPO",
    "build_hpo_from_name",
]



# codes in this file are reproduced from https://github.com/microsoft/nni with some changes.
import torch
import torch.nn as nn
import torch.nn.functional as F
from .base import BaseNAS
from ..space import (
    BaseSpace,
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingInputChoice,
    PathSamplingLayerChoice,
    apply_fixed_architecture,
)
from tqdm import tqdm
import numpy as np
import logging

LOGGER = logging.getLogger(__name__)

from .ea_utils import Individual, UniformSampler, MutationSampler
import collections
import random


class Evolution:
    """
    Algorithm for regularized evolution (i.e. aging evolution).
    Follows "Algorithm 1" in Real et al. "Regularized Evolution for Image Classifier Architecture Search".
    
    Parameters
    ----------
    optimize_mode : str
        Can be one of "maximize" and "minimize". Default: maximize.
    population_size : int
        The number of individuals to keep in the population. Default: 100.
    cycles : int
        The number of cycles (trials) the algorithm should run for. Default: 20000.
    sample_size : int
        The number of individuals that should participate in each tournament. Default: 25.
    mutation_prob : float
        Probability that mutation happens in each dim. Default: 0.05
    """

    def __init__(
        self,
        optimize_mode="maximize",
        population_size=100,
        sample_size=25,
        cycles=20000,
        mutation_prob=0.05,
        disable_progress=False,
    ):
        assert optimize_mode in ["maximize", "minimize"]
        assert sample_size < population_size
        self.optimize_mode = optimize_mode
        self.population_size = population_size
        self.sample_size = sample_size
        self.cycles = cycles
        self.mutation_prob = mutation_prob
        self.disable_progress = disable_progress
        self._worst = (
            float("-inf") if self.optimize_mode == "maximize" else float("inf")
        )

        self._success_count = 0
        self._population = collections.deque()
        self._running_models = []
        self._polling_interval = 2.0
        self._history = []

    def best_parent(self, sample_size=None):
        """get the config of the best parent 
        """
        samples = [p for p in self._population]  # copy population
        random.shuffle(samples)
        if sample_size is not None:
            samples = list(samples)[:sample_size]
        if self.optimize_mode == "maximize":
            parent = max(samples, key=lambda sample: sample.y)
        else:
            parent = min(samples, key=lambda sample: sample.y)
        return parent.x

    def _prepare(self):
        self.uniform = UniformSampler(self.nas_modules)
        self.mutation = MutationSampler(self.nas_modules, self.mutation_prob)

    def _get_metric(self, config):
        for name, module in self.nas_modules:
            module.sampled = config[name]
        # todo: this may be computational expensive
        # model=self.model.parse_model(config,self.device)
        with torch.no_grad():
            metric, loss = self.estimator.infer(self.model, self.dataset, mask="val")
        return metric["acc"]

    def search(self, space: BaseSpace, nas_modules, dset, estimator, device):
        self.model = space
        self.dataset = dset
        self.estimator = estimator
        self.nas_modules = nas_modules
        self.device = device

        self._prepare()
        LOGGER.info("Initializing the first population.")
        with tqdm(range(self.population_size), disable=self.disable_progress) as bar:
            for i in bar:
                config = self.uniform.resample()
                metric = self._get_metric(config)
                individual = Individual(config, metric)
                # LOGGER.debug('Individual created: %s', str(individual))
                self._population.append(individual)
                self._history.append(individual)
                bar.set_postfix(
                    metric=metric,
                    max=max(x.y for x in self._population),
                    min=min(x.y for x in self._population),
                )

        LOGGER.info("Running mutations.")
        with tqdm(range(self.cycles), disable=self.disable_progress) as bar:
            for i in bar:
                parent = self.best_parent(self.sample_size)
                config = self.mutation.resample(parent)
                metric = self._get_metric(config)  # todo : add aging factor
                individual = Individual(config, metric)
                LOGGER.debug("Individual created: %s", str(individual))
                self._population.append(individual)
                self._history.append(individual)
                if len(self._population) > self.population_size:
                    self._population.popleft()
                bar.set_postfix(
                    metric=metric,
                    max_h=max(x.y for x in self._history),
                    max=max(x.y for x in self._population),
                    min=min(x.y for x in self._population),
                )

        # todo: origin is best in history | or the population may need to be retrained
        self._history.sort(key=lambda x: x.y)
        # best=self.best_parent()
        if self.optimize_mode == "maximize":
            best = self._history[-1].x
        else:
            best = self._history[0].x
        return best


class Spos(BaseNAS):
    """
    SPOS trainer.

    Parameters
    ----------
    n_warmup : int
        Number of epochs for training super network.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    
    Other parameters see Evolution
    """

    def __init__(
        self,
        n_warmup=1000,
        grad_clip=5.0,
        disable_progress=False,
        optimize_mode="maximize",
        population_size=100,
        sample_size=25,
        cycles=20000,
        mutation_prob=0.05,
        device="cuda",
    ):
        super().__init__(device)
        self.model_lr = 5e-3
        self.model_wd = 5e-4
        self.n_warmup = n_warmup
        self.disable_progress = disable_progress
        self.grad_clip = grad_clip
        self.optimize_mode = optimize_mode
        self.population_size = population_size
        self.sample_size = sample_size
        self.cycles = cycles
        self.mutation_prob = mutation_prob

    def _prepare(self):
        # replace choice
        self.nas_modules = []
        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)
        # to device
        self.model = self.model.to(self.device)
        self.model_optim = torch.optim.Adam(
            self.model.parameters(), lr=self.model_lr, weight_decay=self.model_wd
        )
        # controller
        self.controller = UniformSampler(self.nas_modules)

        # Evolution
        self.evolve = Evolution(
            optimize_mode="maximize",
            population_size=self.population_size,
            sample_size=self.sample_size,
            cycles=self.cycles,
            mutation_prob=self.mutation_prob,
            disable_progress=self.disable_progress,
        )

    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset
        self.estimator = estimator

        self._prepare()
        self._train()  # train using uniform sampling
        self._search()  # search using evolutionary algorithm

        selection = self.export()

        # here may sample N , retrain N ,and get best
        print(selection)
        return space.parse_model(selection)

    def _search(self):
        self.best_config = self.evolve.search(
            self.model, self.nas_modules, self.dataset, self.estimator, self.device,
        )

    def _train(self):
        with tqdm(range(self.n_warmup), disable=self.disable_progress) as bar:
            for i in bar:
                acc, l1 = self._train_one_epoch(i)
                with torch.no_grad():
                    val_acc, val_loss = self._infer("val")
                bar.set_postfix(
                    loss=l1.item(),
                    acc=acc["acc"],
                    val_acc=val_acc["acc"],
                    val_loss=val_loss.item(),
                )

    def _train_one_epoch(self, epoch):
        self.model.train()
        self.model_optim.zero_grad()
        self._resample()  # uniform sampling
        metric, loss = self._infer(mask="train")
        loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip)
        self.model_optim.step()
        return metric, loss

    def _resample(self):
        result = self.controller.resample()
        for name, module in self.nas_modules:
            module.sampled = result[name]

    def export(self):
        return self.best_config

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.model, self.dataset, mask=mask)
        return metric, loss



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-10-02 08:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Algorithm",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("status", models.CharField(max_length=128)),
                ("created_time", models.DateTimeField(auto_now_add=True)),
                ("updated_time", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Study",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                ("study_configuration", models.TextField()),
                ("algorithm", models.CharField(max_length=128)),
                ("status", models.CharField(max_length=128)),
                ("created_time", models.DateTimeField(auto_now_add=True)),
                ("updated_time", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Trial",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("study_name", models.CharField(max_length=128)),
                ("name", models.CharField(max_length=128)),
                ("parameter_values", models.TextField(blank=True, null=True)),
                ("objective_value", models.FloatField(blank=True, null=True)),
                ("status", models.CharField(max_length=128)),
                ("created_time", models.DateTimeField(auto_now_add=True)),
                ("updated_time", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="TrialMetric",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("trial_id", models.IntegerField()),
                ("training_step", models.IntegerField(blank=True, null=True)),
                ("objective_value", models.FloatField(blank=True, null=True)),
                ("created_time", models.DateTimeField(auto_now_add=True)),
                ("updated_time", models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



from .random_search import RandomSearch
from .darts import Darts
from .rl import RL
from .graphnas_rl import GraphNasRL
from .enas import Enas
from .spos import Spos
from .grna import GRNA
from .gasso import Gasso
from .graces import Graces
from .gauss import Gauss
from .autogt import Autogt



from .message_passing import MessagePassing
from .gcn_conv import GCNConv
from .cheb_conv import ChebConv
from .sage_conv import SAGEConv
from .gat_conv import GATConv
from .gin_conv import GINConv, GINEConv
from .arma_conv import ARMAConv
from .edge_conv import EdgeConv, DynamicEdgeConv
from .ops import ZeroConv, LinearConv, SkipConv
import torch


def gnn_map(gnn_name, in_dim, out_dim, concat=False, bias=True):
    """

    :param gnn_name:
    :param in_dim:
    :param out_dim:
    :param concat: for gat, concat multi-head output or not
    :return: GNN model
    """
    norm = True
    if gnn_name == "gat":
        return GATConv(in_dim, out_dim, 1, bias=bias, concat=False, add_self_loops=norm)
    elif gnn_name == "gcn":
        return GCNConv(in_dim, out_dim, add_self_loops=True, normalize=norm)
    elif gnn_name == "gin":
        return GINConv(torch.nn.Linear(in_dim, out_dim))
    elif gnn_name == "cheb":
        return ChebConv(in_dim, out_dim, K=2, bias=bias)
    elif gnn_name == "sage":
        return SAGEConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "gated":
        return GatedGraphConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "arma":
        return ARMAConv(in_dim, out_dim, bias=bias, normalize=norm)
    elif gnn_name == "sg":
        return SGConv(in_dim, out_dim, bias=bias, normalize=norm)
    elif gnn_name == "linear":
        return LinearConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "skip":
        return SkipConv(in_dim, out_dim, bias=bias)
    elif gnn_name == "zero":
        return ZeroConv(in_dim, out_dim, bias=bias)
    else:
        raise ValueError("No such GNN name")


__all__ = [
    "MessagePassing",
    "GCNConv",
    "ChebConv",
    "SAGEConv",
    "GATConv",
    "GINConv",
    "GINEConv",
    "ARMAConv",
    "EdgeConv",
    "DynamicEdgeConv",
    "gnn_map",
]

classes = __all__



from .choice_ordered import (
    apply_fixed_architecture,
    OrderedMutable,
    OrderedLayerChoice,
    OrderedInputChoice,
    get_module_order,
    sort_replaced_module,
    replace_layer_choice,
    replace_input_choice,
)

from .choice_path import PathSamplingLayerChoice, PathSamplingInputChoice
from .choice_darts import DartsLayerChoice, DartsInputChoice



import json
import itertools

from ...suggestion.models import Study
from ...suggestion.models import Trial
from ...suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm


class GridSearchAlgorithm(AbstractSuggestionAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    """

    def get_new_suggestions(self, study, trials=[], number=1):
        """
        Get the new suggested trials with grid search.
        """

        return_trial_list = []

        # study = Study.objects.get(name=study_name)

        study_configuration_json = json.loads(study.study_configuration)
        params = study_configuration_json["params"]
        param_number = len(params)

        # [['8', '16', '32', '64'], ['sgd', 'adagrad', 'adam', 'ftrl'], ['true', 'false']]
        param_values_list = []

        for param in params:

            # Check param type
            if param["type"] == "DOUBLE" or param["type"] == "INTEGER":
                raise Exception("Grid search does not support DOUBLE and INTEGER")

            feasible_point_list = [
                value.strip() for value in param["feasiblePoints"].split(",")
            ]

            param_values_list.append(feasible_point_list)

        # Example: [('8', 'sgd', 'true'), ('8', 'sgd', 'false'), ('8', 'adagrad', 'true'), ('8', 'adagrad', 'false'), ('8', 'adam', 'true'), ('8', 'adam', 'false'), ('8', 'ftrl', 'true'), ('8', 'ftrl', 'false'), ('16', 'sgd', 'true'), ('16', 'sgd', 'false'), ('16', 'adagrad', 'true'), ('16', 'adagrad', 'false'), ('16', 'adam', 'true'), ('16', 'adam', 'false'), ('16', 'ftrl', 'true'), ('16', 'ftrl', 'false'), ('32', 'sgd', 'true'), ('32', 'sgd', 'false'), ('32', 'adagrad', 'true'), ('32', 'adagrad', 'false'), ('32', 'adam', 'true'), ('32', 'adam', 'false'), ('32', 'ftrl', 'true'), ('32', 'ftrl', 'false'), ('64', 'sgd', 'true'), ('64', 'sgd', 'false'), ('64', 'adagrad', 'true'), ('64', 'adagrad', 'false'), ('64', 'adam', 'true'), ('64', 'adam', 'false'), ('64', 'ftrl', 'true'), ('64', 'ftrl', 'false')]
        combination_values_list = list(itertools.product(*param_values_list))

        # Example: [{"hidden2": "8", "optimizer": "sgd", "batch_normalization": "true"}, ......]
        all_combination_values_json = []

        for combination_values in combination_values_list:

            combination_values_json = {}

            # Example: (u'8', u'sgd', u'true')
            for i in range(param_number):
                # Example: "sgd"
                combination_values_json[
                    params[i]["parameterName"]
                ] = combination_values[i]

            all_combination_values_json.append(combination_values_json)

        all_combination_number = len(all_combination_values_json)

        # Compute how many grid search params have been allocated
        # allocated_trials = Trial.objects.filter(study_name=study_name)
        allocated_trials = trials
        return_trials_start_index = len(allocated_trials)

        if return_trials_start_index > all_combination_number:
            return_trials_start_index = 0
        elif return_trials_start_index + number > all_combination_number:
            return_trials_start_index = all_combination_number - number

        for i in range(number):
            trial = Trial.create(study.name, "GridSearchTrial")
            trial.parameter_values = json.dumps(
                all_combination_values_json[return_trials_start_index + i]
            )
            # trial.save()
            return_trial_list.append(trial)

        return return_trial_list



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from torch.optim.lr_scheduler import _LRScheduler


class PolynomialDecayLR(_LRScheduler):

    def __init__(self, optimizer, warmup_updates, tot_updates, lr, end_lr, power, last_epoch=-1, verbose=False):
        self.warmup_updates = warmup_updates
        self.tot_updates = tot_updates
        self.lr = lr
        self.end_lr = end_lr
        self.power = power
        super(PolynomialDecayLR, self).__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if self._step_count <= self.warmup_updates:
            self.warmup_factor = self._step_count / float(self.warmup_updates)
            lr = self.warmup_factor * self.lr
        elif self._step_count >= self.tot_updates:
            lr = self.end_lr
        else:
            warmup = self.warmup_updates
            lr_range = self.lr - self.end_lr
            pct_remaining = 1 - (self._step_count - warmup) / (
                self.tot_updates - warmup
            )
            lr = lr_range * pct_remaining ** (self.power) + self.end_lr

        return [lr for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        assert False



import torch
from torch import nn
import torch.nn.functional as F


def act_map(act):
    if act == "linear":
        return lambda x: x
    elif act == "elu":
        return F.elu
    elif act == "sigmoid":
        return torch.sigmoid
    elif act == "tanh":
        return torch.tanh
    elif act == "relu":
        return torch.nn.functional.relu
    elif act == "relu6":
        return torch.nn.functional.relu6
    elif act == "softplus":
        return torch.nn.functional.softplus
    elif act == "leaky_relu":
        return torch.nn.functional.leaky_relu
    else:
        raise Exception("wrong activate function")


class LambdaModule(nn.Module):
    def __init__(self, lambd):
        super().__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.lambd)


class StrModule(nn.Module):
    def __init__(self, lambd):
        super().__init__()
        self.str = lambd

    def forward(self, *args, **kwargs):
        return self.str

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.str)


def act_map_nn(act):
    return LambdaModule(act_map(act))


def map_nn(l):
    return [StrModule(x) for x in l]


from autogllight.utils.backend import is_dgl

if is_dgl():
    from .operation_dgl import gnn_map, GeoLayer
else:
    from .operation_pyg import gnn_map, GeoLayer



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import json
import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

from ...suggestion.models import Study
from ...suggestion.models import Trial
from ...suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm
from ...suggestion.algorithm.random_search import RandomSearchAlgorithm


class BayesianOptimization(AbstractSuggestionAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    """

    def get_new_suggestions(self, study, trials=[], number=1):
        # TODO: Only support returning one trial
        number = 1

        # Get study and completed data
        # study = Study.objects.get(name=study_name)
        # completed_trials = Trial.objects.filter(
        #    study_name=study_name, status="Completed")
        completed_trials = [i for i in trials if i.status == "Completed"]
        study_configuration_json = json.loads(study.study_configuration)
        random_init_trial_number = study_configuration_json.get("randomInitTrials", 3)
        params = study_configuration_json["params"]
        study_goal = study_configuration_json["goal"]

        # Use random search if it has less dataset
        if len(completed_trials) < random_init_trial_number:
            randomSearchAlgorithm = RandomSearchAlgorithm()
            return_trials = randomSearchAlgorithm.get_new_suggestions(
                study, trials, number
            )
            return return_trials

        # Construct the map of name and scope to compute gaussian process
        acquisition_function_kappa = 5

        # Example: {'x': (-4, 4), 'y': (-3, 3)}
        # name_scope_map = {}
        # Construct the list with only scope, Example: [(40, 400)]
        bounds = []

        for param in params:

            if param["type"] == "DOUBLE" or param["type"] == "INTEGER":
                min_value = param["minValue"]
                max_value = param["maxValue"]
                # name_scope_map[param["parameterName"]] = (min_value, max_value)
                bounds.append((min_value, max_value))

            elif param["type"] == "DISCRETE":
                feasible_points_string = param["feasiblePoints"]
                feasible_points = [
                    float(value.strip()) for value in feasible_points_string.split(",")
                ]
                for feasible_point in feasible_points:
                    parameter_name = "{}_{}".format(
                        param["parameterName"], feasible_point
                    )
                    # name_scope_map[parameter_name] = (0, 1)
                    bounds.append((0, 1))

            elif param["type"] == "CATEGORICAL":
                feasible_points_string = param["feasiblePoints"]
                feasible_points = [
                    value.strip() for value in feasible_points_string.split(",")
                ]
                for feasible_point in feasible_points:
                    parameter_name = "{}_{}".format(
                        param["parameterName"], feasible_point
                    )
                    # name_scope_map[parameter_name] = (0, 1)
                    bounds.append((0, 1))

        # Make sure it is numpy ndarry
        bounds = np.asarray(bounds)

        # Construct data to train gaussian process, Example: [[50], [150], [250]]
        init_points = []
        # Example: [0.6, 0.8, 0.6]
        init_labels = []

        # Construct train data with completed trials
        for trial in completed_trials:
            # Example: {"learning_rate": 0.01, "optimizer": "ftrl"}
            parameter_values_json = json.loads(trial.parameter_values)

            # Example: [0.01, "ftrl"]
            instance_features = []
            instance_label = trial.objective_value

            for param in params:

                if param["type"] == "DOUBLE" or param["type"] == "INTEGER":
                    instance_feature = parameter_values_json[param["parameterName"]]
                    instance_features.append(instance_feature)

                elif param["type"] == "DISCRETE":
                    feasible_points_string = param["feasiblePoints"]
                    feasible_points = [
                        float(value.strip())
                        for value in feasible_points_string.split(",")
                    ]
                    parameter_value = parameter_values_json[param["parameterName"]]
                    for feasible_point in feasible_points:
                        if feasible_point == parameter_value:
                            instance_features.append(1)
                        else:
                            instance_features.append(0)

                elif param["type"] == "CATEGORICAL":
                    feasible_points_string = param["feasiblePoints"]
                    # Example: ["sgd", "adagrad", "adam", "ftrl"]
                    feasible_points = [
                        value.strip() for value in feasible_points_string.split(",")
                    ]
                    # Example: "ftrl"
                    parameter_value = parameter_values_json[param["parameterName"]]
                    for feasible_point in feasible_points:
                        if feasible_point == parameter_value:
                            instance_features.append(1)
                        else:
                            instance_features.append(0)

            init_points.append(instance_features)
            init_labels.append(instance_label)

        # Example: ndarray([[ 50], [150], [250]])
        train_features = np.asarray(init_points)
        # Example: ndarray([0.6, 0.8, 0.6])
        train_labels = np.asarray(init_labels)
        # current_max_label = train_labels.max()

        # Train with gaussian process
        gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            n_restarts_optimizer=25,
        )

        gp.fit(train_features, train_labels)

        # Example: [[-3.66909025, -0.84486644], [-1.93270006, -0.95367483], [1.36095631, 0.61358525], ...], shape is [100000, 2]
        x_tries = np.random.uniform(
            bounds[:, 0], bounds[:, 1], size=(100000, bounds.shape[0])
        )

        mean, std = gp.predict(x_tries, return_std=True)

        # Confidence bound criteria
        acquisition_fucntion_values = mean + acquisition_function_kappa * std

        # x_max = x_tries[acquisition_fucntion_values.argmax()]
        # tobe
        # x_max = x_tries[acquisition_fucntion_values.argmin()]

        if study_goal == "MAXIMIZE":
            x_max = x_tries[acquisition_fucntion_values.argmax()]
            # max_acquision_fucntion_value = acquisition_fucntion_values.max()
        elif study_goal == "MINIMIZE":
            x_max = x_tries[acquisition_fucntion_values.argmin()]
            # max_acquision_fucntion_value = acquisition_fucntion_values.min()
        else:
            # TODO: Throw the error
            x_max = []

        # Example: [3993.864683994805, 44.15441513231316]
        x_max = np.clip(x_max, bounds[:, 0], bounds[:, 1])
        print("Current max acquision function choose: {}".format(x_max))

        # Example: {"hidden2": 3993.864683994805, "hidden1": 44.15441513231316}
        suggested_parameter_values_json = {}

        index = 0
        """
    Construct the suggested params according to the result of gaussian process
    # Example prior result: [0.1, 0.5, 0.3, 0.9]
    # Example param scope: {"learning_rate": (0.01, 0.5), "hidden1": (40, 400), "optimizer_sgd": (0, 1), "optimizer_ftrl": (0, 1)}
    for key in bound_dict.keys():
      parameter_values_json[key] = x_max[index]
      index += 1
    """

        for param in params:

            if param["type"] == "DOUBLE":
                suggested_parameter_values_json[param["parameterName"]] = x_max[index]
                index += 1

            elif param["type"] == "INTEGER":
                suggested_parameter_values_json[param["parameterName"]] = int(
                    x_max[index]
                )
                index += 1

            elif param["type"] == "DISCRETE":
                feasible_points_string = param["feasiblePoints"]
                feasible_points = [
                    float(value.strip()) for value in feasible_points_string.split(",")
                ]

                # Find the max value of these and get its string
                current_max = x_max[index]
                suggested_parameter_value = feasible_points[0]

                for feasible_point in feasible_points:
                    if x_max[index] > current_max:
                        current_max = x_max[index]
                        suggested_parameter_value = feasible_point
                    index += 1

                suggested_parameter_values_json[
                    param["parameterName"]
                ] = suggested_parameter_value

            elif param["type"] == "CATEGORICAL":
                feasible_points_string = param["feasiblePoints"]
                # Example: ["sgd", "adagrad", "adam", "ftrl"]
                feasible_points = [
                    value.strip() for value in feasible_points_string.split(",")
                ]

                # Find the max value of these and get its string
                current_max = x_max[index]
                suggested_parameter_value = feasible_points[0]

                for feasible_point in feasible_points:
                    if x_max[index] > current_max:
                        current_max = x_max[index]
                        suggested_parameter_value = feasible_point
                    index += 1

                suggested_parameter_values_json[
                    param["parameterName"]
                ] = suggested_parameter_value

        return_trial = Trial.create(study.name, "BayesianOptimizationTrial")
        return_trial.parameter_values = json.dumps(suggested_parameter_values_json)

        return [return_trial]


class BayesianOptimizationDemo(object):
    def test_function2(self, x, y):
        return -(x ** 2) - (y - 1) ** 2 + 1

    def test_function(self, x, y):
        # The best x is 2
        return np.exp(-((x - 2) ** 2)) + np.exp(-((x - 6) ** 2) / 10) + 1 / (x ** 2 + 1)

    def test_bayes_optimizaion(self):
        print("Start bayesian optimization")

        # 1. Initialize parameters
        acquisition_function_kappa = 5
        init_point_number = 3
        iteration_number = 3
        iteration_index = 0
        train_features = []
        train_labels = []

        gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            n_restarts_optimizer=25,
        )

        bound_dict = {"x": (-4, 4), "y": (-3, 3)}
        # Example: [[-4,  4], [-3,  3]]
        bounds = []
        for key in bound_dict.keys():
            bounds.append(bound_dict[key])
        # Example: ndarray([[-4,  4], [-3,  3]])
        bounds = np.asarray(bounds)

        # 2. Get init random samples
        # Example: array([-3.66909025, -1.93270006, 1.36095631])
        init_xs = np.random.uniform(-4, 4, size=init_point_number)
        # Example: array([-0.84486644, -0.95367483, 0.61358525])
        init_ys = np.random.uniform(-3, 3, size=init_point_number)

        # Example: [[-3.66909025, -0.84486644], [-1.93270006, -0.95367483], [1.36095631, 0.61358525]]
        init_points = []
        for i in range(init_point_number):
            init_points.append([init_xs[i], init_ys[i]])

        # Example: [-4.4555402320291684, -7.9016857176523114]
        init_labels = []
        for point in init_points:
            init_labels.append(self.test_function(point[0], point[1]))

        # 3. GP compute the prior
        train_features = np.asarray(init_points)
        train_labels = np.asarray(init_labels)
        current_max_label = train_labels.max()

        gp.fit(train_features, train_labels)

        # 4. Acquision function computes the max value
        # Example: [[-3.66909025, -0.84486644], [-1.93270006, -0.95367483], [1.36095631, 0.61358525], ...], shape is [100000, 2]
        x_tries = np.random.uniform(
            bounds[:, 0], bounds[:, 1], size=(100000, bounds.shape[0])
        )

        mean, std = gp.predict(x_tries, return_std=True)
        # Confidence bound criteria
        acquisition_fucntion_values = mean + acquisition_function_kappa * std
        x_max = x_tries[acquisition_fucntion_values.argmax()]
        max_acquision_fucntion_value = acquisition_fucntion_values.max()

        x_max = np.clip(x_max, bounds[:, 0], bounds[:, 1])
        print("Current max acquision function choose: {}".format(x_max))

        for i in range(iteration_number):
            iteration_index += 1

            # 5. Choose the best and compute to add in train dataset
            train_features = np.vstack((train_features, x_max.reshape((1, -1))))
            train_labels = np.append(
                train_labels, self.test_function(x_max[0], x_max[1])
            )

            # 6. Re-compute gaussian process and acquistion function
            gp.fit(train_features, train_labels)

            # Update maximum value
            if train_labels[-1] > current_max_label:
                current_max_label = train_labels[-1]
                print("Get the better parameters!")

            x_tries = np.random.uniform(
                bounds[:, 0], bounds[:, 1], size=(100000, bounds.shape[0])
            )

            mean, std = gp.predict(x_tries, return_std=True)
            acquisition_fucntion_values = mean + acquisition_function_kappa * std
            x_max = x_tries[acquisition_fucntion_values.argmax()]
            max_acquision_fucntion_value = acquisition_fucntion_values.max()

            x_max = np.clip(x_max, bounds[:, 0], bounds[:, 1])
            print(
                "Max label: {}, current label: {}, acquision function choose: {}".format(
                    current_max_label, train_labels[-1], x_max
                )
            )
